from __future__ import annotations

import json
import os
import pathlib
import shutil
import unittest.mock
from typing import Any, Callable, Dict, Optional, Union, cast
from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

import pytest
import requests

import great_expectations as gx
from great_expectations.core.yaml_handler import YAMLHandler
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.types.base import DataContextConfig
from great_expectations.data_context.util import file_relative_path
from great_expectations.datasource.fluent.interfaces import Datasource

yaml = YAMLHandler()

# No longer used but keeping around for tests
USAGE_STATISTICS_QA_URL = (
    "https://qa.stats.greatexpectations.io/great_expectations/v1/usage_statistics"
)


@pytest.fixture()
def data_context_without_config_variables_filepath_configured(tmp_path_factory):
    # This data_context is *manually* created to have the config we want, vs created with DataContext.create  # noqa: E501 # FIXME CoP
    project_path = str(tmp_path_factory.mktemp("data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    asset_config_path = os.path.join(context_path, "expectations")  # noqa: PTH118 # FIXME CoP

    create_data_context_files(
        context_path,
        asset_config_path,
        ge_config_fixture_filename="great_expectations_basic_without_config_variables_filepath.yml",
        config_variables_fixture_filename=None,
    )

    return gx.get_context(context_root_dir=context_path)


@pytest.fixture()
def data_context_with_variables_in_config(tmp_path_factory, monkeypatch):
    monkeypatch.setenv("FOO", "BAR")
    monkeypatch.setenv("REPLACE_ME_ESCAPED_ENV", "ive_been_$--replaced")
    # This data_context is *manually* created to have the config we want, vs created with DataContext.create  # noqa: E501 # FIXME CoP
    project_path = str(tmp_path_factory.mktemp("data_context"))
    context_path = os.path.join(project_path, FileDataContext.GX_DIR)  # noqa: PTH118 # FIXME CoP
    asset_config_path = os.path.join(context_path, "expectations")  # noqa: PTH118 # FIXME CoP

    create_data_context_files(
        context_path,
        asset_config_path,
        ge_config_fixture_filename="great_expectations_basic_with_variables.yml",
        config_variables_fixture_filename="config_variables.yml",
    )

    return gx.get_context(context_root_dir=context_path)


def create_data_context_files(
    context_path,
    asset_config_path,
    ge_config_fixture_filename,
    config_variables_fixture_filename=None,
):
    if config_variables_fixture_filename:
        os.makedirs(context_path, exist_ok=True)  # noqa: PTH103 # FIXME CoP
        os.makedirs(  # noqa: PTH103 # FIXME CoP
            os.path.join(context_path, "uncommitted"),  # noqa: PTH118 # FIXME CoP
            exist_ok=True,
        )
        copy_relative_path(
            f"../test_fixtures/{config_variables_fixture_filename}",
            str(
                os.path.join(  # noqa: PTH118 # FIXME CoP
                    context_path, "uncommitted/config_variables.yml"
                )
            ),
        )
        copy_relative_path(
            f"../test_fixtures/{ge_config_fixture_filename}",
            str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
        )
    else:
        os.makedirs(context_path, exist_ok=True)  # noqa: PTH103 # FIXME CoP
        copy_relative_path(
            f"../test_fixtures/{ge_config_fixture_filename}",
            str(os.path.join(context_path, FileDataContext.GX_YML)),  # noqa: PTH118 # FIXME CoP
        )
    create_common_data_context_files(context_path, asset_config_path)


def create_common_data_context_files(context_path, asset_config_path):
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(  # noqa: PTH118 # FIXME CoP
            asset_config_path, "mydatasource/mygenerator/my_dag_node"
        ),
        exist_ok=True,
    )
    copy_relative_path(
        "../test_fixtures/expectation_suites/parameterized_expectation_suite_fixture.json",
        os.path.join(  # noqa: PTH118 # FIXME CoP
            asset_config_path, "mydatasource/mygenerator/my_dag_node/default.json"
        ),
    )
    os.makedirs(  # noqa: PTH103 # FIXME CoP
        os.path.join(context_path, "plugins"),  # noqa: PTH118 # FIXME CoP
        exist_ok=True,
    )


def copy_relative_path(relative_src, dest):
    shutil.copy(file_relative_path(__file__, relative_src), dest)


@pytest.fixture
def basic_data_context_config():
    return DataContextConfig(
        **{
            "commented_map": {},
            "config_version": 2,
            "plugins_directory": "plugins/",
            "validation_results_store_name": "does_not_have_to_be_real",
            "expectations_store_name": "expectations_store",
            "checkpoint_store_name": "checkpoint_store",
            "config_variables_file_path": "uncommitted/config_variables.yml",
            "stores": {
                "expectations_store": {
                    "class_name": "ExpectationsStore",
                    "store_backend": {
                        "class_name": "TupleFilesystemStoreBackend",
                        "base_directory": "expectations/",
                    },
                },
                "checkpoint_store": {
                    "class_name": "CheckpointStore",
                    "store_backend": {
                        "class_name": "TupleFilesystemStoreBackend",
                        "base_directory": "checkpoints/",
                    },
                },
            },
            "data_docs_sites": {},
            "analytics_enabled": True,
            "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        }
    )


@pytest.fixture
def basic_data_context_config_dict(basic_data_context_config):
    """Wrapper fixture to transform `basic_data_context_config` to a json dict"""
    return basic_data_context_config.to_json_dict()


@pytest.fixture
def conn_string_password():
    """Returns a stable password for mocking connection strings"""
    return "not_a_password"


@pytest.fixture
def conn_string_with_embedded_password(conn_string_password):
    """
    A mock connection string with the `conn_string_password` fixture embedded.
    """
    return f"redshift+psycopg2://no_user:{conn_string_password}@111.11.1.1:1111/foo"


@pytest.fixture
def data_context_config_with_datasources(conn_string_password):
    return DataContextConfig(
        **{
            "commented_map": {},
            "config_version": 2,
            "plugins_directory": "plugins/",
            "validation_results_store_name": "does_not_have_to_be_real",
            "expectations_store_name": "expectations_store",
            "checkpoint_store_name": "checkpoint_store",
            "config_variables_file_path": "uncommitted/config_variables.yml",
            "fluent_datasources": {
                "Datasource 2: Postgres": {
                    "type": "pandas_filesystem",
                    "base_directory": "/path/to/trip_data",
                }
            },
            "stores": {
                "expectations_store": {
                    "class_name": "ExpectationsStore",
                    "store_backend": {
                        "class_name": "TupleFilesystemStoreBackend",
                        "base_directory": "expectations/",
                    },
                },
                "checkpoint_store": {
                    "class_name": "CheckpointStore",
                    "store_backend": {
                        "class_name": "TupleFilesystemStoreBackend",
                        "base_directory": "checkpoints/",
                    },
                },
            },
            "data_docs_sites": {},
            "analytics_enabled": True,
            "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        }
    )


@pytest.fixture
def data_context_config_dict_with_datasources(data_context_config_with_datasources):
    """Wrapper fixture to transform `data_context_config_with_datasources` to a json dict"""
    return data_context_config_with_datasources.to_json_dict()


@pytest.fixture
def data_context_config_with_cloud_backed_stores(ge_cloud_access_token):
    org_id = "a34595b2-267e-4469-b18f-774e65dc556f"
    return DataContextConfig(
        **{
            "commented_map": {},
            "config_version": 2,
            "plugins_directory": "plugins/",
            "validation_results_store_name": "does_not_have_to_be_real",
            "expectations_store_name": "expectations_store",
            "config_variables_file_path": "uncommitted/config_variables.yml",
            "stores": {
                "default_checkpoint_store": {
                    "class_name": "CheckpointStore",
                    "store_backend": {
                        "class_name": GXCloudStoreBackend.__name__,
                        "ge_cloud_base_url": "http://foo/bar/",
                        "ge_cloud_credentials": {
                            "access_token": ge_cloud_access_token,
                            "organization_id": org_id,
                        },
                        "ge_cloud_resource_type": "checkpoint",
                        "suppress_store_backend_id": True,
                    },
                },
                "default_expectations_store": {
                    "class_name": "ExpectationsStore",
                    "store_backend": {
                        "class_name": GXCloudStoreBackend.__name__,
                        "ge_cloud_base_url": "http://foo/bar/",
                        "ge_cloud_credentials": {
                            "access_token": ge_cloud_access_token,
                            "organization_id": org_id,
                        },
                        "ge_cloud_resource_type": "expectation_suite",
                        "suppress_store_backend_id": True,
                    },
                },
                "default_validation_results_store": {
                    "class_name": "ValidationResultsStore",
                    "store_backend": {
                        "class_name": GXCloudStoreBackend.__name__,
                        "ge_cloud_base_url": "http://foo/bar/",
                        "ge_cloud_credentials": {
                            "access_token": ge_cloud_access_token,
                            "organization_id": org_id,
                        },
                        "ge_cloud_resource_type": "validation_result",
                        "suppress_store_backend_id": True,
                    },
                },
            },
            "data_docs_sites": {},
            "analytics_enabled": True,
            "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
        }
    )


@pytest.fixture
def data_context_config_dict_with_cloud_backed_stores(
    data_context_config_with_cloud_backed_stores,
):
    """Wrapper fixture to transform `data_context_config_with_cloud_backed_stores` to a json dict"""
    return data_context_config_with_cloud_backed_stores.to_json_dict()


@pytest.fixture
def ge_cloud_runtime_base_url():
    return "https://api.dev.greatexpectations.io/runtime"


@pytest.fixture
def ge_cloud_runtime_organization_id():
    return "a8a35168-68d5-4366-90ae-00647463d37e"


@pytest.fixture
def datasource_name() -> str:
    return "my_first_datasource"


@pytest.fixture
def datasource_store_name() -> str:
    return "datasource_store"


@pytest.fixture
def fake_datasource_id() -> str:
    return "aaa7cfdd-4aa4-4f3d-a979-fe2ea5203cbf"


@pytest.fixture
def fake_data_connector_id() -> str:
    return "0c08e6ba-8ed9-4715-a179-da2f08aab13e"


JSONData = Union[Dict[str, Any]]
RequestError = Union[requests.exceptions.HTTPError, requests.exceptions.Timeout]


class MockResponse:
    # TODO: GG 08232022 update signature to accept arbitrary content types
    def __init__(
        self,
        json_data: JSONData,
        status_code: int,
        headers: Optional[Dict[str, str]] = None,
        exc_to_raise: Optional[RequestError] = None,
    ) -> None:
        self._json_data = json_data
        self.status_code = status_code
        self.headers = headers or {"content-type": "application/json" if json_data else "text/html"}
        self._exc_to_raise = exc_to_raise

    def json(self):
        if self.headers.get("content-type") == "application/json":
            return self._json_data
        raise json.JSONDecodeError("Uh oh - check content-type", "foobar", 1)

    def raise_for_status(self):
        if self._exc_to_raise:
            raise self._exc_to_raise
        if self.status_code >= 400:
            raise requests.exceptions.HTTPError(f"Mock {self.status_code} HTTPError", response=self)

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return f"<Response [{self.status_code}]>"


@pytest.fixture
def mock_response_factory() -> Callable[[JSONData, int, Optional[RequestError]], MockResponse]:
    def _make_mock_response(
        json_data: JSONData,
        status_code: int,
        exc_to_raise: Optional[RequestError] = None,
    ) -> MockResponse:
        return MockResponse(json_data=json_data, status_code=status_code, exc_to_raise=exc_to_raise)

    return _make_mock_response


def basic_fluent_datasource_config() -> dict:
    return {
        "type": "pandas_filesystem",
        "name": "my_fluent_pandas_filesystem_datasource",
        "assets": [
            {
                "name": "my_csv",
                "type": "csv",
            }
        ],
        "base_directory": pathlib.PosixPath("/path/to/trip_data"),
    }


def basic_fluent_datasource() -> Datasource:
    context = gx.get_context(mode="ephemeral")
    datasource = context.data_sources.add_pandas_filesystem(
        name="pandas_filesystem",
        base_directory="/path/to/trip_data",  # type: ignore [arg-type]
    )
    datasource.add_csv_asset(name="my_csv")
    return datasource


@pytest.fixture
def fluent_datasource_config() -> dict:
    return basic_fluent_datasource_config()


@pytest.fixture
def mock_http_unavailable(mock_response_factory: Callable):
    """Mock all request http calls to return a 503 Unavailable response."""

    def mocked_response(*args, **kwargs):
        return MockResponse(
            {"code": 503, "detail": "API is unavailable"},
            503,
        )

    # should have been able to do this by mocking `requests.request` but this didn't work
    with unittest.mock.patch.multiple(
        "requests.Session",
        autospec=True,
        get=unittest.mock.DEFAULT,
        post=unittest.mock.DEFAULT,
        put=unittest.mock.DEFAULT,
        patch=unittest.mock.DEFAULT,
        delete=unittest.mock.DEFAULT,
    ) as mock_requests:
        for name, mock in cast("Dict[str, Mock]", mock_requests).items():
            mock.side_effect = mocked_response
            print(f"Mocking `requests.{name}` with `{mocked_response.__name__}()`")

        yield mock_requests


@pytest.fixture
def checkpoint_config() -> dict:
    checkpoint_config = {
        "name": "oss_test_checkpoint",
        "expectation_suite_name": "oss_test_expectation_suite",
        "validations": [
            {
                "name": None,
                "id": None,
                "expectation_suite_name": "taxi.demo_pass",
                "expectation_suite_id": None,
                "batch_request": None,
            },
            {
                "name": None,
                "id": None,
                "expectation_suite_name": None,
                "expectation_suite_id": None,
                "batch_request": {
                    "datasource_name": "oss_test_datasource",
                    "data_connector_name": "oss_test_data_connector",
                    "data_asset_name": "users",
                },
            },
        ],
        "action_list": [
            {
                "action": {"class_name": "StoreValidationResultAction"},
                "name": "store_validation_result",
            },
        ],
    }
    return checkpoint_config


@pytest.fixture
def mocked_datasource_post_response(
    mock_response_factory: Callable,
    fake_datasource_id: str,
) -> Callable[[], MockResponse]:
    def _mocked_post_response(*args, **kwargs):
        return mock_response_factory(
            {
                "data": {
                    "id": fake_datasource_id,
                }
            },
            201,
        )

    return _mocked_post_response



from typing import Literal, Optional

from great_expectations.compatibility.pydantic import StrictStr
from great_expectations.expectations.model_field_types import ConditionParser
from great_expectations.metrics.metric import Metric, _MetricResult


class MultiColumnMetric(Metric[_MetricResult], kw_only=True):
    column_list: list[str]
    row_condition: Optional[StrictStr] = None
    condition_parser: Optional[ConditionParser] = None
    ignore_row_if: Literal["all_values_are_missing", "any_value_is_missing", "never"] = (
        "all_values_are_missing"
    )



from __future__ import annotations

import datetime
import hashlib
import logging
import pickle
from functools import partial
from io import BytesIO
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Tuple,
    Union,
    cast,
    overload,
)

import pandas as pd

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import aws, azure, google
from great_expectations.compatibility.sqlalchemy_and_pandas import (
    execute_pandas_reader_fn,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch import BatchMarkers
from great_expectations.core.batch_spec import (
    AzureBatchSpec,
    BatchSpec,
    FabricBatchSpec,
    GCSBatchSpec,
    PandasBatchSpec,
    PandasBatchSpecProtocol,
    PathBatchSpec,
    RuntimeDataBatchSpec,
    S3BatchSpec,
)
from great_expectations.core.metric_domain_types import (
    MetricDomainTypes,  # noqa: TC001 # FIXME CoP
)
from great_expectations.core.util import AzureUrl, GCSUrl, S3Url, sniff_s3_compression
from great_expectations.execution_engine import ExecutionEngine
from great_expectations.execution_engine.execution_engine import (
    PartitionDomainKwargs,  # noqa: TC001 # FIXME CoP
)
from great_expectations.execution_engine.pandas_batch_data import PandasBatchData
from great_expectations.execution_engine.partition_and_sample.pandas_data_partitioner import (
    PandasDataPartitioner,
)
from great_expectations.execution_engine.partition_and_sample.pandas_data_sampler import (
    PandasDataSampler,
)
from great_expectations.expectations.model_field_types import CONDITION_PARSER_PANDAS

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from great_expectations.validator.metric_configuration import MetricConfigurationID

logger = logging.getLogger(__name__)


HASH_THRESHOLD = 1e9

DataFrameFactoryFn: TypeAlias = Callable[..., pd.DataFrame]


class PandasExecutionEngine(ExecutionEngine):
    """PandasExecutionEngine instantiates the ExecutionEngine API to support computations using Pandas.

    Constructor builds a PandasExecutionEngine, using provided configuration options.

    Args:
        *args: Positional arguments for configuring PandasExecutionEngine
        **kwargs: Keyword arguments for configuring PandasExecutionEngine

    For example:
    ```python
        execution_engine: ExecutionEngine = PandasExecutionEngine(batch_data_dict={batch.id: batch.data})
    ```

    --ge-feature-maturity-info--

        id: validation_engine_pandas
        title: Validation Engine - Pandas
        icon:
        short_description: Use Pandas DataFrame to validate data
        description: Use Pandas DataFrame to validate data
        how_to_guide_url:
        maturity: Production
        maturity_details:
            api_stability: Stable
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: N/A -> see relevant Datasource evaluation
            documentation_completeness: Complete
            bug_risk: Low
            expectation_completeness: Complete

    --ge-feature-maturity-info--
    """  # noqa: E501 # FIXME CoP

    recognized_batch_spec_defaults = {
        "reader_method",
        "reader_options",
    }

    def __init__(self, *args, **kwargs) -> None:
        self.discard_subset_failing_expectations = kwargs.pop(
            "discard_subset_failing_expectations", False
        )
        boto3_options: Dict[str, dict] = kwargs.pop("boto3_options", {})
        azure_options: Dict[str, dict] = kwargs.pop("azure_options", {})
        gcs_options: Dict[str, dict] = kwargs.pop("gcs_options", {})

        # Instantiate cloud provider clients as None at first.
        # They will be instantiated if/when passed cloud-specific in BatchSpec is passed in
        self._s3 = None
        self._azure: azure.BlobServiceClient | None = None
        self._gcs = None

        super().__init__(*args, **kwargs)

        self._config.update(
            {
                "discard_subset_failing_expectations": self.discard_subset_failing_expectations,
                "boto3_options": boto3_options,
                "azure_options": azure_options,
                "gcs_options": gcs_options,
            }
        )

        self._data_partitioner = PandasDataPartitioner()
        self._data_sampler = PandasDataSampler()

    def _instantiate_azure_client(self) -> None:
        self._azure = None
        if azure.BlobServiceClient:  # type: ignore[truthy-function] # False if NotImported
            azure_options = self.config.get("azure_options", {})
            try:
                if "conn_str" in azure_options:
                    self._azure = azure.BlobServiceClient.from_connection_string(**azure_options)
                else:
                    self._azure = azure.BlobServiceClient(**azure_options)
            except (TypeError, AttributeError):
                # If exception occurs, then "self._azure = None" remains in effect.
                pass

    def _instantiate_s3_client(self) -> None:
        # Try initializing cloud provider client. If unsuccessful, we'll die here.
        boto3_options = self.config.get("boto3_options", {})
        self._s3 = aws.boto3.client("s3", **boto3_options)

    def _instantiate_gcs_client(self) -> None:
        """
        Helper method for instantiating GCS client when GCSBatchSpec is passed in.

        The method accounts for 3 ways that a GCS connection can be configured:
            1. setting an environment variable, which is typically GOOGLE_APPLICATION_CREDENTIALS
            2. passing in explicit credentials via gcs_options
            3. running Great Expectations from within a GCP container, at which you would be able to create a Client
                without passing in an additional environment variable or explicit credentials
        """  # noqa: E501 # FIXME CoP
        gcs_options = self.config.get("gcs_options", {})
        try:
            credentials = None  # If configured with gcloud CLI / env vars
            if "filename" in gcs_options:
                filename = gcs_options.pop("filename")
                credentials = google.service_account.Credentials.from_service_account_file(
                    filename=filename
                )
            elif "info" in gcs_options:
                info = gcs_options.pop("info")
                credentials = google.service_account.Credentials.from_service_account_info(
                    info=info
                )
            self._gcs = google.storage.Client(credentials=credentials, **gcs_options)
        # This exception handling causes a TypeError if google dependency not installed
        except (TypeError, AttributeError, google.DefaultCredentialsError):
            self._gcs = None

    @override
    def configure_validator(self, validator) -> None:
        super().configure_validator(validator)
        validator.expose_dataframe_methods = True

    @override
    def load_batch_data(
        self,
        batch_id: str,
        batch_data: Union[PandasBatchData, pd.DataFrame],  # type: ignore[override] # FIXME CoP
    ) -> None:
        if isinstance(batch_data, pd.DataFrame):
            batch_data = PandasBatchData(self, batch_data)
        elif not isinstance(batch_data, PandasBatchData):
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "PandasExecutionEngine requires batch data that is either a DataFrame or a PandasBatchData object"  # noqa: E501 # FIXME CoP
            )

        super().load_batch_data(batch_id=batch_id, batch_data=batch_data)

    @override
    def get_batch_data_and_markers(  # noqa: C901, PLR0912, PLR0915 # FIXME CoP
        self, batch_spec: BatchSpec | PandasBatchSpecProtocol
    ) -> Tuple[PandasBatchData, BatchMarkers]:  # batch_data
        # We need to build a batch_markers to be used in the dataframe
        batch_markers = BatchMarkers(
            {
                "ge_load_time": datetime.datetime.now(datetime.timezone.utc).strftime(
                    "%Y%m%dT%H%M%S.%fZ"
                )
            }
        )

        batch_data: Any
        if isinstance(batch_spec, RuntimeDataBatchSpec):
            # batch_data != None is already checked when RuntimeDataBatchSpec is instantiated
            batch_data = batch_spec.batch_data
            if isinstance(batch_data, str):
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""PandasExecutionEngine has been passed a string type batch_data, "{batch_data}", which is illegal.  Please check your config.
"""  # noqa: E501 # FIXME CoP
                )

            if isinstance(batch_spec.batch_data, pd.DataFrame):
                df = batch_spec.batch_data
            elif isinstance(batch_spec.batch_data, PandasBatchData):
                df = batch_spec.batch_data.dataframe
            else:
                raise ValueError(  # noqa: TRY003, TRY004 # FIXME CoP
                    "RuntimeDataBatchSpec must provide a Pandas DataFrame or PandasBatchData object."  # noqa: E501 # FIXME CoP
                )

            batch_spec.batch_data = "PandasDataFrame"

        elif isinstance(batch_spec, S3BatchSpec):
            if self._s3 is None:
                self._instantiate_s3_client()
            s3_engine = self._s3
            try:
                reader_method: str = batch_spec.reader_method
                reader_options: dict = batch_spec.reader_options or {}
                path: str = batch_spec.path
                s3_url = S3Url(path)
                if "compression" not in reader_options:
                    inferred_compression_param = sniff_s3_compression(s3_url)
                    if inferred_compression_param is not None:
                        reader_options["compression"] = inferred_compression_param
                if s3_engine:
                    s3_object: dict = s3_engine.get_object(Bucket=s3_url.bucket, Key=s3_url.key)
            except (
                aws.exceptions.ParamValidationError,
                aws.exceptions.ClientError,
            ) as error:
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""PandasExecutionEngine encountered the following error while trying to read data from S3 Bucket: {error}"""  # noqa: E501 # FIXME CoP
                )
            logger.debug(f"Fetching s3 object. Bucket: {s3_url.bucket} Key: {s3_url.key}")
            reader_fn: DataFrameFactoryFn = self._get_reader_fn(reader_method, s3_url.key)
            buf = BytesIO(s3_object["Body"].read())  # type: ignore[possibly-undefined] # FIXME
            buf.seek(0)
            df = reader_fn(buf, **reader_options)

        elif isinstance(batch_spec, AzureBatchSpec):
            if self._azure is None:
                self._instantiate_azure_client()
            # if we were not able to instantiate Azure client, then raise error
            if self._azure is None:
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    """PandasExecutionEngine has been passed a AzureBatchSpec,
                        but the ExecutionEngine does not have an Azure client configured. Please check your config."""  # noqa: E501 # FIXME CoP
                )
            azure_engine = self._azure
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options or {}
            path = batch_spec.path
            azure_url = AzureUrl(path)
            blob_client = azure_engine.get_blob_client(
                container=azure_url.container, blob=azure_url.blob
            )
            azure_object = blob_client.download_blob()
            logger.debug(
                f"Fetching Azure blob. Container: {azure_url.container} Blob: {azure_url.blob}"
            )
            reader_fn = self._get_reader_fn(reader_method, azure_url.blob)
            buf = BytesIO(azure_object.readall())
            buf.seek(0)
            df = reader_fn(buf, **reader_options)

        elif isinstance(batch_spec, GCSBatchSpec):
            if self._gcs is None:
                self._instantiate_gcs_client()
            # if we were not able to instantiate GCS client, then raise error
            if self._gcs is None:
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    """PandasExecutionEngine has been passed a GCSBatchSpec,
                        but the ExecutionEngine does not have an GCS client configured. Please check your config."""  # noqa: E501 # FIXME CoP
                )
            gcs_engine = self._gcs
            gcs_url = GCSUrl(batch_spec.path)
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options or {}
            try:
                gcs_bucket = gcs_engine.get_bucket(gcs_url.bucket)
                gcs_blob = gcs_bucket.blob(gcs_url.blob)
                logger.debug(f"Fetching GCS blob. Bucket: {gcs_url.bucket} Blob: {gcs_url.blob}")
            except google.GoogleAPIError as error:
                raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""PandasExecutionEngine encountered the following error while trying to read data from GCS \
Bucket: {error}"""  # noqa: E501 # FIXME CoP
                )
            reader_fn = self._get_reader_fn(reader_method, gcs_url.blob)
            buf = BytesIO(gcs_blob.download_as_bytes())
            buf.seek(0)
            df = reader_fn(buf, **reader_options)

        # Experimental datasources will go down this code path
        elif isinstance(batch_spec, PathBatchSpec):
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options
            path = batch_spec.path
            reader_fn = self._get_reader_fn(reader_method, path)
            df = reader_fn(path, **reader_options)

        elif isinstance(batch_spec, PandasBatchSpec):
            reader_method = batch_spec.reader_method
            reader_options = batch_spec.reader_options
            reader_fn = self._get_reader_fn(reader_method)
            reader_fn_result: pd.DataFrame | list[pd.DataFrame] = execute_pandas_reader_fn(
                reader_fn, reader_options
            )
            if isinstance(reader_fn_result, list):
                if len(reader_fn_result) > 1:
                    raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                        "Pandas reader method must return a single DataFrame, "
                        f'but "{reader_method}" returned {len(reader_fn_result)} DataFrames.'
                    )
                else:
                    df = reader_fn_result[0]
            else:
                df = reader_fn_result

        elif isinstance(batch_spec, FabricBatchSpec):
            reader_fn = batch_spec.get_reader_function()
            df = reader_fn(**batch_spec.reader_options)

        else:
            raise gx_exceptions.BatchSpecError(  # noqa: TRY003 # FIXME CoP
                f"""batch_spec must be of type RuntimeDataBatchSpec, PandasBatchSpec, PathBatchSpec, S3BatchSpec, AzureBatchSpec or FabricBatchSpec \
not {batch_spec.__class__.__name__}"""  # noqa: E501 # FIXME CoP
            )

        df = self._apply_partitioning_and_sampling_methods(batch_spec, df)  # type: ignore[arg-type] # FIXME CoP
        if df.memory_usage().sum() < HASH_THRESHOLD:
            batch_markers["pandas_data_fingerprint"] = hash_pandas_dataframe(df)

        typed_batch_data = PandasBatchData(execution_engine=self, dataframe=df)

        return typed_batch_data, batch_markers

    def _apply_partitioning_and_sampling_methods(
        self,
        batch_spec: BatchSpec | PandasBatchSpecProtocol,
        batch_data: PandasBatchData,
    ):
        # partitioning and sampling not supported for FabricBatchSpec
        if isinstance(batch_spec, BatchSpec):
            partitioner_method_name: Optional[str] = batch_spec.get("partitioner_method")
            if partitioner_method_name:
                partitioner_fn: Callable = self._data_partitioner.get_partitioner_method(
                    partitioner_method_name
                )
                partitioner_kwargs: dict = batch_spec.get("partitioner_kwargs") or {}
                batch_data = partitioner_fn(batch_data, **partitioner_kwargs)

            sampler_method_name: Optional[str] = batch_spec.get("sampling_method")
            if sampler_method_name:
                sampling_fn: Callable = self._data_sampler.get_sampler_method(sampler_method_name)
                batch_data = sampling_fn(batch_data, batch_spec)

        return batch_data

    @property
    def dataframe(self) -> pd.DataFrame:
        """Tests whether or not a Batch has been loaded. If the loaded batch does not exist, raises a
        ValueError Exception
        """  # noqa: E501 # FIXME CoP
        # Changed to is None because was breaking prior
        if self.batch_manager.active_batch_data is None:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "Batch has not been loaded - please run load_batch_data() to load a batch."
            )

        return cast("PandasBatchData", self.batch_manager.active_batch_data).dataframe

    # NOTE Abe 20201105: Any reason this shouldn't be a private method?
    @staticmethod
    def guess_reader_method_from_path(path: str):  # noqa: C901, PLR0911 # FIXME CoP
        """Helper method for deciding which reader to use to read in a certain path.

        Args:
            path (str): the to use to guess

        Returns:
            ReaderMethod to use for the filepath

        """
        path = path.lower()
        if path.endswith(".csv") or path.endswith(".tsv"):
            return {"reader_method": "read_csv"}
        elif path.endswith(".parquet") or path.endswith(".parq") or path.endswith(".pqt"):
            return {"reader_method": "read_parquet"}
        elif path.endswith(".xlsx") or path.endswith(".xls"):
            return {"reader_method": "read_excel"}
        elif path.endswith(".json"):
            return {"reader_method": "read_json"}
        elif path.endswith(".pkl"):
            return {"reader_method": "read_pickle"}
        elif path.endswith(".feather"):
            return {"reader_method": "read_feather"}
        elif path.endswith(".csv.gz") or path.endswith(".tsv.gz"):
            return {
                "reader_method": "read_csv",
                "reader_options": {"compression": "gzip"},
            }
        elif path.endswith(".sas7bdat") or path.endswith(".xpt"):
            return {"reader_method": "read_sas"}

        else:
            raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                f'Unable to determine reader method from path: "{path}".'
            )

    @overload
    def _get_reader_fn(
        self, reader_method: str = ..., path: Optional[str] = ...
    ) -> DataFrameFactoryFn: ...

    @overload
    def _get_reader_fn(self, reader_method: None = ..., path: str = ...) -> DataFrameFactoryFn: ...

    def _get_reader_fn(
        self, reader_method: Optional[str] = None, path: Optional[str] = None
    ) -> DataFrameFactoryFn:
        """Static helper for parsing reader types. If reader_method is not provided, path will be used to guess the
        correct reader_method.

        Args:
            reader_method (str): the name of the reader method to use, if available.
            path (str): the path used to guess

        Returns:
            ReaderMethod to use for the filepath

        """  # noqa: E501 # FIXME CoP
        if reader_method is None and path is None:
            raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                "Unable to determine pandas reader function without reader_method or path."
            )

        reader_options = {}
        if reader_method is None:
            path_guess = self.guess_reader_method_from_path(path)  # type: ignore[arg-type] # see overload
            reader_method = path_guess["reader_method"]
            reader_options = path_guess.get(
                "reader_options"
            )  # This may not be there; use None in that case

        try:
            reader_fn = getattr(pd, reader_method)
            if reader_options:
                reader_fn = partial(reader_fn, **reader_options)
            return reader_fn
        except AttributeError:
            raise gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                f'Unable to find reader_method "{reader_method}" in pandas.'
            )

    @override
    def resolve_metric_bundle(self, metric_fn_bundle) -> dict[MetricConfigurationID, Any]:
        """Resolve a bundle of metrics with the same compute Domain as part of a single trip to the compute engine."""  # noqa: E501 # FIXME CoP
        return {}  # This is NO-OP for "PandasExecutionEngine" (no bundling for direct execution computational backend).  # noqa: E501 # FIXME CoP

    @override
    def get_domain_records(  # noqa: C901, PLR0912 # FIXME CoP
        self,
        domain_kwargs: dict,
    ) -> pd.DataFrame:
        """Uses the given Domain kwargs (which include row_condition, condition_parser, and ignore_row_if directives) to obtain and/or query a Batch of data.

        Args:
            domain_kwargs (dict) - A dictionary consisting of the Domain kwargs specifying which data to obtain

        Returns:
            A DataFrame (the data on which to compute returned in the format of a Pandas DataFrame)
        """  # noqa: E501 # FIXME CoP
        table = domain_kwargs.get("table", None)
        if table:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "PandasExecutionEngine does not currently support multiple named tables."
            )

        batch_id = domain_kwargs.get("batch_id")
        if batch_id is None:
            # We allow no batch id specified if there is only one batch
            if self.batch_manager.active_batch_data_id is not None:
                data = cast("PandasBatchData", self.batch_manager.active_batch_data).dataframe
            else:
                raise gx_exceptions.ValidationError(  # noqa: TRY003 # FIXME CoP
                    "No batch is specified, but could not identify a loaded batch."
                )
        else:  # noqa: PLR5501 # FIXME CoP
            if batch_id in self.batch_manager.batch_data_cache:
                data = cast(
                    "PandasBatchData", self.batch_manager.batch_data_cache[batch_id]
                ).dataframe
            else:
                raise gx_exceptions.ValidationError(  # noqa: TRY003 # FIXME CoP
                    f"Unable to find batch with batch_id {batch_id}"
                )

        # Filtering by row condition.
        row_condition = domain_kwargs.get("row_condition", None)
        if row_condition:
            condition_parser = domain_kwargs.get("condition_parser", None)

            if condition_parser == CONDITION_PARSER_PANDAS:
                data = data.query(row_condition, parser=condition_parser)
            else:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "condition_parser for Pandas is required when setting a row_condition."
                )

        if "column" in domain_kwargs:
            return data

        if (
            "column_A" in domain_kwargs
            and "column_B" in domain_kwargs
            and "ignore_row_if" in domain_kwargs
        ):
            # noinspection PyPep8Naming
            column_A_name = domain_kwargs["column_A"]
            # noinspection PyPep8Naming
            column_B_name = domain_kwargs["column_B"]

            ignore_row_if = domain_kwargs["ignore_row_if"]
            if ignore_row_if == "both_values_are_missing":
                data = data.dropna(
                    axis=0,
                    how="all",
                    subset=[column_A_name, column_B_name],
                )
            elif ignore_row_if == "either_value_is_missing":
                data = data.dropna(
                    axis=0,
                    how="any",
                    subset=[column_A_name, column_B_name],
                )
            else:  # noqa: PLR5501 # FIXME CoP
                if ignore_row_if != "neither":
                    raise ValueError(f'Unrecognized value of ignore_row_if ("{ignore_row_if}").')  # noqa: TRY003 # FIXME CoP

            return data

        if "column_list" in domain_kwargs and "ignore_row_if" in domain_kwargs:
            column_list = domain_kwargs["column_list"]

            ignore_row_if = domain_kwargs["ignore_row_if"]
            if ignore_row_if == "all_values_are_missing":
                data = data.dropna(
                    axis=0,
                    how="all",
                    subset=column_list,
                )
            elif ignore_row_if == "any_value_is_missing":
                data = data.dropna(
                    axis=0,
                    how="any",
                    subset=column_list,
                )
            else:  # noqa: PLR5501 # FIXME CoP
                if ignore_row_if != "never":
                    raise ValueError(f'Unrecognized value of ignore_row_if ("{ignore_row_if}").')  # noqa: TRY003 # FIXME CoP

            return data

        return data

    @override
    def get_compute_domain(
        self,
        domain_kwargs: dict,
        domain_type: Union[str, MetricDomainTypes],
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> Tuple[pd.DataFrame, dict, dict]:
        """Uses the given Domain kwargs (which include row_condition, condition_parser, and ignore_row_if directives) to obtain and/or query a batch.

        Returns in the format of a Pandas DataFrame along with Domain arguments required for computing.  If the Domain \
        is a single column, this is added to 'accessor Domain kwargs' and used for later access.

        Args:
            domain_kwargs (dict): a dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type (str or MetricDomainTypes): an Enum value indicating which metric Domain the user would like \
            to be using, or a corresponding string value representing it.  String types include "column", \
            "column_pair", "table", and "other".  Enum types include capitalized versions of these from the class \
            MetricDomainTypes.
            accessor_keys (str iterable): keys that are part of the compute Domain but should be ignored when \
            describing the Domain and simply transferred with their associated values into accessor_domain_kwargs.

        Returns:
            A tuple including:
              - a DataFrame (the data on which to compute)
              - a dictionary of compute_domain_kwargs, describing the DataFrame
              - a dictionary of accessor_domain_kwargs, describing any accessors needed to
                identify the Domain within the compute domain
        """  # noqa: E501 # FIXME CoP
        table: str = domain_kwargs.get("table", None)
        if table:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "PandasExecutionEngine does not currently support multiple named tables."
            )

        data: pd.DataFrame = self.get_domain_records(domain_kwargs=domain_kwargs)

        partition_domain_kwargs: PartitionDomainKwargs = self._partition_domain_kwargs(
            domain_kwargs, domain_type, accessor_keys
        )

        return data, partition_domain_kwargs.compute, partition_domain_kwargs.accessor


def hash_pandas_dataframe(df):
    try:
        obj = pd.util.hash_pandas_object(df, index=True).values
    except TypeError:
        # In case of facing unhashable objects (like dict), use pickle
        obj = pickle.dumps(df, pickle.HIGHEST_PROTOCOL)

    return hashlib.md5(obj).hexdigest()



# isort:skip_file

import pathlib

from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    Datasource,
    Sorter,
    BatchMetadata,
    GxDatasourceWarning,
    GxContextWarning,
    TestConnectionError,
)
from great_expectations.datasource.fluent.invalid_datasource import (
    InvalidAsset,
    InvalidDatasource,
    GxInvalidDatasourceWarning,
)

# Now that DataAsset has both been defined, we need to
# provide it to the BatchDefinition pydantic model.
from great_expectations.core.batch_definition import BatchDefinition

BatchDefinition.update_forward_refs(DataAsset=DataAsset)


from great_expectations.datasource.fluent.alloy_datasource import (
    AlloyDatasource,
)
from great_expectations.datasource.fluent.aurora_datasource import (
    AuroraDatasource,
)
from great_expectations.datasource.fluent.batch_request import (
    BatchRequest,
    BatchParameters,
)
from great_expectations.datasource.fluent.bigquery_datasource import (
    BigQueryDatasource,
)
from great_expectations.datasource.fluent.citus_datasource import (
    CitusDatasource,
)
from great_expectations.datasource.fluent.pandas_datasource import (
    PandasDatasource,
    _PandasDatasource,
)
from great_expectations.datasource.fluent.pandas_file_path_datasource import (
    _PandasFilePathDatasource,
)
from great_expectations.datasource.fluent.pandas_filesystem_datasource import (
    PandasFilesystemDatasource,
)
from great_expectations.datasource.fluent.pandas_dbfs_datasource import (
    PandasDBFSDatasource,
)
from great_expectations.datasource.fluent.pandas_s3_datasource import (
    PandasS3Datasource,
)
from great_expectations.datasource.fluent.pandas_google_cloud_storage_datasource import (
    PandasGoogleCloudStorageDatasource,
)
from great_expectations.datasource.fluent.pandas_azure_blob_storage_datasource import (
    PandasAzureBlobStorageDatasource,
)
from great_expectations.datasource.fluent.fabric import FabricPowerBIDatasource
from great_expectations.datasource.fluent.postgres_datasource import (
    PostgresDatasource,
)
from great_expectations.datasource.fluent.neon_datasource import (
    NeonDatasource,
)
from great_expectations.datasource.fluent.redshift_datasource import RedshiftDatasource
from great_expectations.datasource.fluent.spark_datasource import (
    _SparkDatasource,
)
from great_expectations.datasource.fluent.spark_datasource import (
    SparkDatasource,
)
from great_expectations.datasource.fluent.spark_file_path_datasource import (
    _SparkFilePathDatasource,
)
from great_expectations.datasource.fluent.spark_filesystem_datasource import (
    SparkFilesystemDatasource,
)
from great_expectations.datasource.fluent.spark_dbfs_datasource import (
    SparkDBFSDatasource,
)
from great_expectations.datasource.fluent.spark_s3_datasource import (
    SparkS3Datasource,
)
from great_expectations.datasource.fluent.spark_google_cloud_storage_datasource import (
    SparkGoogleCloudStorageDatasource,
)
from great_expectations.datasource.fluent.spark_azure_blob_storage_datasource import (
    SparkAzureBlobStorageDatasource,
)
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource
from great_expectations.datasource.fluent.sqlite_datasource import (
    SqliteDatasource,
)
from great_expectations.datasource.fluent.databricks_sql_datasource import (
    DatabricksSQLDatasource,
)
from great_expectations.datasource.fluent.snowflake_datasource import (
    SnowflakeDatasource,
)


_PANDAS_SCHEMA_VERSION: str = (
    "1.5.3"  # this is the version schemas we generated for. Update as needed
)
_SCHEMAS_DIR = pathlib.Path(__file__).parent / "schemas"



from __future__ import annotations

import logging
from typing import Tuple

from great_expectations.expectations.metrics.metric_provider import MetricProvider

logger = logging.getLogger(__name__)


class TableMetricProvider(MetricProvider):
    """Base class for all Table Metrics, which define metrics to be calculated across a complete table.

    An example of this is `table.column.unique`,
    which verifies that a table contains unique, non-duplicate columns.

    Args:
        metric_name (str): A name identifying the metric. Metric Name must be globally unique in
            a great_expectations installation.
        domain_keys (tuple): A tuple of the keys used to determine the domain of the metric.
        value_keys (tuple): A tuple of the keys used to determine the value of the metric.

    In some cases, subclasses of MetricProvider, such as TableMetricProvider, will already
    have correct values that may simply be inherited by Metric classes.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/expectations/custom_expectations_lp
    """  # noqa: E501 # FIXME CoP

    domain_keys: Tuple[str, ...] = (
        "batch_id",
        "table",
        "row_condition",
        "condition_parser",
    )



from typing import Final

import pandas as pd
import pytest
from contrib.experimental.great_expectations_experimental.expectations.expect_column_values_to_be_present_in_other_table import (  # noqa: E501 # FIXME CoP
    ExpectColumnValuesToBePresentInOtherTable,  # needed for expectation registration
)

from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.data_context.util import file_relative_path
from great_expectations.datasource.fluent import SqliteDatasource

DB_PATH: Final[str] = file_relative_path(
    __file__, "../../test_sets/referential_integrity_dataset.db"
)


@pytest.fixture
def referential_integrity_db(sa):
    """Create a sqlite database with 3 tables: order_table_1, order_table_2, and customer_table. We only run this once to create the database."""  # noqa: E501 # FIXME CoP
    sqlite_engine = sa.create_engine(f"sqlite:///{DB_PATH}")
    order_table_1 = pd.DataFrame(
        {
            "ORDER_ID": ["aaa", "bbb", "ccc"],
            "CUSTOMER_ID": [1, 1, 3],
        }
    )
    order_table_2 = pd.DataFrame(
        {
            "ORDER_ID": ["aaa", "bbb", "ccc"],
            "CUSTOMER_ID": [1, 5, 6],
        }
    )
    customer_table = pd.DataFrame(
        {
            "CUSTOMER_ID": [1, 2, 3],
        }
    )

    add_dataframe_to_db(
        df=order_table_1,
        name="order_table_1",
        con=sqlite_engine,
        index=False,
        if_exists="replace",
    )
    add_dataframe_to_db(
        df=order_table_2,
        name="order_table_2",
        con=sqlite_engine,
        index=False,
        if_exists="replace",
    )
    add_dataframe_to_db(
        df=customer_table,
        name="customer_table",
        con=sqlite_engine,
        index=False,
        if_exists="replace",
    )
    return sqlite_engine


@pytest.fixture()
def sqlite_datasource(in_memory_runtime_context) -> SqliteDatasource:
    context = in_memory_runtime_context
    datasource_name = "my_snowflake_datasource"
    return context.data_sources.add_sqlite(
        datasource_name, connection_string=f"sqlite:///{DB_PATH}"
    )


@pytest.mark.sqlite
def test_successful_expectation_run(sqlite_datasource):
    datasource = sqlite_datasource
    asset_name = "order_table_1"
    asset = datasource.add_table_asset(name=asset_name, table_name="order_table_1")
    batch = asset.get_batch(asset.build_batch_request())
    res = batch.validate(
        ExpectColumnValuesToBePresentInOtherTable(
            foreign_key_column="CUSTOMER_ID",
            foreign_table="customer_table",
            foreign_table_key_column="CUSTOMER_ID",
        )
    )
    assert res.success is True


@pytest.mark.sqlite
def test_failed_expectation_run(sqlite_datasource):
    datasource = sqlite_datasource
    asset_name = "order_table_2"
    asset = datasource.add_table_asset(name=asset_name, table_name="order_table_2")
    batch = asset.get_batch(asset.build_batch_request())
    res = batch.validate(
        ExpectColumnValuesToBePresentInOtherTable(
            foreign_key_column="CUSTOMER_ID",
            foreign_table="customer_table",
            foreign_table_key_column="CUSTOMER_ID",
        )
    )
    assert res.success is False
    assert res["result"]["observed_value"] == "2 missing values."
    assert res["result"]["unexpected_index_list"] == [
        {"CUSTOMER_ID": 5},
        {"CUSTOMER_ID": 6},
    ]


@pytest.mark.sqlite
def test_configuration_invalid_column_name(sqlite_datasource):
    """What does this test do, and why?

    This is testing default behavior of `batch.validate()` which catches Exception information
    and places it in `exception_info`. Here we check that the exception message contains the text we expect
    """  # noqa: E501 # FIXME CoP
    datasource = sqlite_datasource
    asset_name = "order_table_2"
    asset = datasource.add_table_asset(name=asset_name, table_name="order_table_2")
    batch = asset.get_batch(asset.build_batch_request())
    res = batch.validate(
        ExpectColumnValuesToBePresentInOtherTable(
            foreign_key_column="I_DONT_EXIST",
            foreign_table="customer_table",
            foreign_table_key_column="CUSTOMER_ID",
        ),
    )

    assert res.success is False
    for k, v in res["exception_info"].items():
        assert v["raised_exception"] is True
        assert "no such column: a.I_DONT_EXIST" in v["exception_message"]


@pytest.mark.unit
def test_template_dict_creation():
    expectation = ExpectColumnValuesToBePresentInOtherTable(
        foreign_key_column="CUSTOMER_ID",
        foreign_table="customer_table",
        foreign_table_key_column="CUSTOMER_ID",
    )
    assert expectation.template_dict == {
        "foreign_key_column": "CUSTOMER_ID",
        "foreign_table": "customer_table",
        "foreign_table_key_column": "CUSTOMER_ID",
    }



from __future__ import annotations

import json

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResultMeta,
)
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)


@pytest.mark.unit
def test_expectation_validation_result_describe_returns_expected_description():
    # arrange
    evr = ExpectationValidationResult(
        success=False,
        expectation_config=gxe.ExpectColumnValuesToBeBetween(
            column="passenger_count",
            min_value=0,
            max_value=6,
            notes="Per the TLC data dictionary, this is a driver-submitted value (historically between 0 to 6)",  # noqa: E501 # FIXME CoP
        ).configuration,
        result={
            "element_count": 100000,
            "unexpected_count": 1,
            "unexpected_percent": 0.001,
            "partial_unexpected_list": [7.0],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.001,
            "unexpected_percent_nonmissing": 0.001,
            "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
            "partial_unexpected_index_list": [48422],
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
    )
    # act
    description = evr.describe()
    # assert
    assert description == json.dumps(
        {
            "expectation_type": "expect_column_values_to_be_between",
            "success": False,
            "kwargs": {"column": "passenger_count", "min_value": 0.0, "max_value": 6.0},
            "result": {
                "element_count": 100000,
                "unexpected_count": 1,
                "unexpected_percent": 0.001,
                "partial_unexpected_list": [7.0],
                "missing_count": 0,
                "missing_percent": 0.0,
                "unexpected_percent_total": 0.001,
                "unexpected_percent_nonmissing": 0.001,
                "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                "partial_unexpected_index_list": [48422],
            },
        },
        indent=4,
    )


@pytest.mark.unit
def test_expectation_validation_result_describe_returns_expected_description_with_null_values():
    # It's unclear if an ExpectationValidationResult can ever be valid without an Expectation
    # or a result, but since it's typed that way we test it
    # arrange
    evr = ExpectationValidationResult(
        success=True,
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
    )
    # act
    description = evr.describe()
    # assert
    assert description == json.dumps(
        {
            "expectation_type": None,
            "success": True,
            "kwargs": None,
            "result": {},
        },
        indent=4,
    )


@pytest.mark.unit
def test_expectation_validation_result_describe_returns_expected_description_with_exception():
    # arrange
    evr = ExpectationValidationResult(
        success=False,
        expectation_config=gxe.ExpectColumnValuesToBeBetween(
            column="passenger_count",
            min_value=0,
            max_value=6,
            notes="Per the TLC data dictionary, this is a driver-submitted value (historically between 0 to 6)",  # noqa: E501 # FIXME CoP
        ).configuration,
        result={
            "element_count": 100000,
            "unexpected_count": 1,
            "unexpected_percent": 0.001,
            "partial_unexpected_list": [7.0],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.001,
            "unexpected_percent_nonmissing": 0.001,
            "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
            "partial_unexpected_index_list": [48422],
        },
        exception_info={
            "raised_exception": True,
            "exception_traceback": "Traceback (most recent call last): something went wrong",
            "exception_message": "Helpful message here",
        },
    )
    # act
    description = evr.describe()
    # assert
    assert description == json.dumps(
        {
            "expectation_type": "expect_column_values_to_be_between",
            "success": False,
            "kwargs": {"column": "passenger_count", "min_value": 0.0, "max_value": 6.0},
            "result": {
                "element_count": 100000,
                "unexpected_count": 1,
                "unexpected_percent": 0.001,
                "partial_unexpected_list": [7.0],
                "missing_count": 0,
                "missing_percent": 0.0,
                "unexpected_percent_total": 0.001,
                "unexpected_percent_nonmissing": 0.001,
                "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                "partial_unexpected_index_list": [48422],
            },
            "exception_info": {
                "raised_exception": True,
                "exception_traceback": "Traceback (most recent call last): something went wrong",
                "exception_message": "Helpful message here",
            },
        },
        indent=4,
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "validation_result_url",
    [
        "https://app.greatexpectations.io/organizations/my-org/data-assets/6f6d390b-a52b-41d1-b5c0-a1d57a6b4618/validations/expectation-suites/a0af0eb5-90ab-4219-ab60-482eee0a8b32/results/e77ce5e4-b71b-4f86-9c3b-f82385aab660",
        None,
    ],
)
def test_expectation_suite_validation_result_returns_expected_shape(
    validation_result_url: str | None,
):
    # arrange
    svr = ExpectationSuiteValidationResult(
        success=True,
        statistics={
            "evaluated_expectations": 2,
            "successful_expectations": 2,
            "unsuccessful_expectations": 0,
            "success_percent": 100.0,
        },
        suite_name="empty_suite",
        results=[
            ExpectationValidationResult(
                **{
                    "meta": {},
                    "success": True,
                    "exception_info": {
                        "raised_exception": False,
                        "exception_traceback": None,
                        "exception_message": None,
                    },
                    "result": {
                        "element_count": 100000,
                        "unexpected_count": 1,
                        "unexpected_percent": 0.001,
                        "partial_unexpected_list": [7.0],
                        "missing_count": 0,
                        "missing_percent": 0.0,
                        "unexpected_percent_total": 0.001,
                        "unexpected_percent_nonmissing": 0.001,
                        "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                        "partial_unexpected_index_list": [48422],
                    },
                    "expectation_config": ExpectationConfiguration(
                        **{
                            "meta": {},
                            "notes": "Per the TLC data dictionary, this is a driver-submitted value (historically between 0 to 6)",  # noqa: E501 # FIXME CoP
                            "id": "9f76d0b5-9d99-4ed9-a269-339b35e60490",
                            "kwargs": {
                                "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                                "mostly": 0.95,
                                "column": "passenger_count",
                                "min_value": 0.0,
                                "max_value": 6.0,
                            },
                            "type": "expect_column_values_to_be_between",
                        }
                    ),
                }
            ),
            ExpectationValidationResult(
                **{
                    "meta": {},
                    "success": True,
                    "exception_info": {
                        "raised_exception": False,
                        "exception_traceback": None,
                        "exception_message": None,
                    },
                    "result": {
                        "element_count": 100000,
                        "unexpected_count": 0,
                        "unexpected_percent": 0.0,
                        "partial_unexpected_list": [],
                        "partial_unexpected_counts": [],
                        "partial_unexpected_index_list": [],
                    },
                    "expectation_config": ExpectationConfiguration(
                        **{
                            "meta": {},
                            "id": "19c0e80c-d676-4b01-a4a3-2a568552d368",
                            "kwargs": {
                                "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                                "column": "trip_distance",
                            },
                            "type": "expect_column_values_to_not_be_null",
                        }
                    ),
                }
            ),
        ],
        result_url=validation_result_url,
    )
    # act
    description = svr.describe()
    # assert
    assert description == json.dumps(
        {
            "success": True,
            "statistics": {
                "evaluated_expectations": 2,
                "successful_expectations": 2,
                "unsuccessful_expectations": 0,
                "success_percent": 100.0,
            },
            "expectations": [
                {
                    "expectation_type": "expect_column_values_to_be_between",
                    "success": True,
                    "kwargs": {
                        "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                        "mostly": 0.95,
                        "column": "passenger_count",
                        "min_value": 0.0,
                        "max_value": 6.0,
                    },
                    "result": {
                        "element_count": 100000,
                        "unexpected_count": 1,
                        "unexpected_percent": 0.001,
                        "partial_unexpected_list": [7.0],
                        "missing_count": 0,
                        "missing_percent": 0.0,
                        "unexpected_percent_total": 0.001,
                        "unexpected_percent_nonmissing": 0.001,
                        "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                        "partial_unexpected_index_list": [48422],
                    },
                },
                {
                    "expectation_type": "expect_column_values_to_not_be_null",
                    "success": True,
                    "kwargs": {
                        "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                        "column": "trip_distance",
                    },
                    "result": {
                        "element_count": 100000,
                        "unexpected_count": 0,
                        "unexpected_percent": 0.0,
                        "partial_unexpected_list": [],
                        "partial_unexpected_counts": [],
                        "partial_unexpected_index_list": [],
                    },
                },
            ],
            "result_url": validation_result_url,
        },
        indent=4,
    )


@pytest.mark.unit
@pytest.mark.parametrize(
    "validation_result_url",
    [
        "https://app.greatexpectations.io/organizations/my-org/data-assets/6f6d390b-a52b-41d1-b5c0-a1d57a6b4618/validations/expectation-suites/a0af0eb5-90ab-4219-ab60-482eee0a8b32/results/e77ce5e4-b71b-4f86-9c3b-f82385aab660",
        None,
    ],
)
def test_expectation_suite_validation_asset_name_access(
    validation_result_url: str | None,
):
    # arrange
    svr = ExpectationSuiteValidationResult(
        meta=ExpectationSuiteValidationResultMeta(
            **{
                "active_batch_definition": {
                    "batch_identifiers": {},
                    "data_asset_name": "taxi_data_1.csv",
                    "data_connector_name": "default_inferred_data_connector_name",
                    "datasource_name": "pandas",
                },
                "batch_markers": {
                    "ge_load_time": "20220727T154327.630107Z",
                    "pandas_data_fingerprint": "c4f929e6d4fab001fedc9e075bf4b612",
                },
                "batch_spec": {"path": "../data/taxi_data_1.csv"},
                "checkpoint_name": "single_validation_checkpoint",
                "expectation_suite_name": "taxi_suite_1",
                "great_expectations_version": "0.15.15",
                "run_id": {
                    "run_name": "20220727-114327-my-run-name-template",
                    "run_time": "2022-07-27T11:43:27.625252+00:00",
                },
                "validation_time": "20220727T154327.701100Z",
            }
        ),
        success=True,
        statistics={
            "evaluated_expectations": 2,
            "successful_expectations": 2,
            "unsuccessful_expectations": 0,
            "success_percent": 100.0,
        },
        suite_name="empty_suite",
        results=[
            ExpectationValidationResult(
                **{
                    "meta": {},
                    "success": True,
                    "exception_info": {
                        "raised_exception": False,
                        "exception_traceback": None,
                        "exception_message": None,
                    },
                    "result": {
                        "element_count": 100000,
                        "unexpected_count": 1,
                        "unexpected_percent": 0.001,
                        "partial_unexpected_list": [7.0],
                        "missing_count": 0,
                        "missing_percent": 0.0,
                        "unexpected_percent_total": 0.001,
                        "unexpected_percent_nonmissing": 0.001,
                        "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
                        "partial_unexpected_index_list": [48422],
                    },
                    "expectation_config": ExpectationConfiguration(
                        **{
                            "meta": {},
                            "notes": "Test notes",
                            "id": "9f76d0b5-9d99-4ed9-a269-339b35e60490",
                            "kwargs": {
                                "batch_id": "default_pandas_datasource-#ephemeral_pandas_asset",
                            },
                            "type": "expect_column_values_to_be_between",
                        }
                    ),
                }
            ),
        ],
        result_url=validation_result_url,
    )

    assert svr.asset_name == "taxi_data_1.csv"


@pytest.mark.unit
def test_render_updates_rendered_content():
    evr = ExpectationValidationResult(
        success=False,
        expectation_config=gxe.ExpectColumnValuesToBeBetween(
            column="passenger_count",
            min_value=0,
            max_value=6,
            notes="Per the TLC data dictionary, this is a driver-submitted value (historically between 0 to 6)",  # noqa: E501 # FIXME CoP
        ).configuration,
        result={
            "element_count": 100000,
            "unexpected_count": 1,
            "unexpected_percent": 0.001,
            "partial_unexpected_list": [7.0],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.001,
            "unexpected_percent_nonmissing": 0.001,
            "partial_unexpected_counts": [{"value": 7.0, "count": 1}],
            "partial_unexpected_index_list": [48422],
        },
    )

    assert evr.rendered_content is None

    evr.render()

    assert evr.rendered_content is not None


class TestSerialization:
    @pytest.mark.unit
    def test_expectation_validation_results_serializes(self) -> None:
        evr = ExpectationValidationResult(
            success=True,
            expectation_config=gxe.ExpectColumnDistinctValuesToEqualSet(
                column="passenger_count",
                value_set=[1, 2],
            ).configuration,
            result={
                "details": {
                    "observed_value": pd.Series({"a": 1, "b": 2, "c": 4}),
                }
            },
        )

        # Ensure the results are serializable.
        as_dict = evr.describe_dict()
        from_describe_dict = json.dumps(as_dict, indent=4)
        from_describe = evr.describe()

        assert from_describe_dict == from_describe
        assert as_dict["result"]["details"]["observed_value"] == [
            {"index": "a", "value": 1},
            {"index": "b", "value": 2},
            {"index": "c", "value": 4},
        ]

    @pytest.mark.unit
    def test_expectation_suite_validation_results_serializes(self) -> None:
        svr = ExpectationSuiteValidationResult(
            success=True,
            statistics={
                "evaluated_expectations": 2,
                "successful_expectations": 2,
                "unsuccessful_expectations": 0,
                "success_percent": 100.0,
            },
            suite_name="whatever",
            results=[
                ExpectationValidationResult(
                    success=True,
                    expectation_config=gxe.ExpectColumnDistinctValuesToEqualSet(
                        column="passenger_count",
                        value_set=[1, 2],
                    ).configuration,
                    result={
                        "details": {
                            "observed_value": pd.Series({"a": 1, "b": 2, "c": 4}),
                        }
                    },
                )
            ],
        )

        # Ensure the results are serializable.
        as_dict = svr.describe_dict()
        from_describe_dict = json.dumps(as_dict, indent=4)
        from_describe = svr.describe()

        assert from_describe_dict == from_describe
        assert as_dict["expectations"][0]["result"]["details"]["observed_value"] == [
            {"index": "a", "value": 1},
            {"index": "b", "value": 2},
            {"index": "c", "value": 4},
        ]


class TestExpectationValidationResultHash:
    @pytest.mark.unit
    def test_hash_consistency_with_equality(self):
        config1 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )
        config2 = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        result1 = ExpectationValidationResult(
            success=True,
            expectation_config=config1,
            result={"observed_value": 100},
            meta={"test": "value"},
            exception_info={"raised_exception": False},
        )

        result2 = ExpectationValidationResult(
            success=True,
            expectation_config=config2,
            result={"observed_value": 100},
            meta={"test": "value"},
            exception_info={"raised_exception": False},
        )

        assert result1 == result2
        assert hash(result1) == hash(result2)

    @pytest.mark.unit
    def test_hash_different_for_different_success(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        result1 = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 100}
        )

        result2 = ExpectationValidationResult(
            success=False, expectation_config=config, result={"observed_value": 100}
        )

        assert result1 != result2
        assert hash(result1) != hash(result2)

    @pytest.mark.unit
    def test_hash_different_for_different_results(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        result1 = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 100}
        )

        result2 = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 200}
        )

        assert result1 != result2
        assert hash(result1) != hash(result2)

    @pytest.mark.unit
    def test_hash_stable_across_runs(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        result = ExpectationValidationResult(
            success=True,
            expectation_config=config,
            result={"observed_value": 100},
            meta={"test": "value"},
            exception_info={"raised_exception": False},
        )

        hash1 = hash(result)
        hash2 = hash(result)
        hash3 = hash(result)

        assert hash1 == hash2 == hash3


class TestExpectationSuiteValidationResultHash:
    @pytest.mark.unit
    def test_hash_consistency_with_equality(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        evr = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 100}
        )

        result1 = ExpectationSuiteValidationResult(
            suite_name="test_suite",
            success=True,
            results=[evr],
            suite_parameters={"param": "value"},
            statistics={"evaluated_expectations": 1},
            meta={"test": "value"},
        )

        result2 = ExpectationSuiteValidationResult(
            suite_name="test_suite",
            success=True,
            results=[evr],
            suite_parameters={"param": "value"},
            statistics={"evaluated_expectations": 1},
            meta={"test": "value"},
        )

        assert result1 == result2
        assert hash(result1) == hash(result2)

    @pytest.mark.unit
    def test_hash_different_for_different_success(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        evr = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 100}
        )

        result1 = ExpectationSuiteValidationResult(
            suite_name="test_suite",
            success=True,
            results=[evr],
            suite_parameters={"param": "value"},
        )

        result2 = ExpectationSuiteValidationResult(
            suite_name="test_suite",
            success=False,
            results=[evr],
            suite_parameters={"param": "value"},
        )

        assert result1 != result2
        assert hash(result1) != hash(result2)

    @pytest.mark.unit
    def test_hash_stable_across_runs(self):
        config = ExpectationConfiguration(
            type="expect_column_values_to_not_be_null", kwargs={"column": "test_column"}
        )

        evr = ExpectationValidationResult(
            success=True, expectation_config=config, result={"observed_value": 100}
        )

        result = ExpectationSuiteValidationResult(
            suite_name="test_suite",
            success=True,
            results=[evr],
            suite_parameters={"param": "value"},
            statistics={"evaluated_expectations": 1},
            meta={"test": "value"},
        )

        hash1 = hash(result)
        hash2 = hash(result)
        hash3 = hash(result)

        assert hash1 == hash2 == hash3



from __future__ import annotations

import abc
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
    Sequence,
)

from great_expectations.core.domain import SemanticDomainTypes
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.experimental.metric_repository.column_filter import ColumnFilter
from great_expectations.experimental.metric_repository.metrics import (
    ColumnMetric,
    MetricException,
    MetricTypes,
    TableMetric,
)
from great_expectations.validator.exception_info import ExceptionInfo
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
    MetricConfigurationID,
)

if TYPE_CHECKING:
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.datasource.fluent import BatchRequest
    from great_expectations.experimental.metric_repository.metrics import Metric
    from great_expectations.validator.metrics_calculator import (
        _AbortedMetricsInfoDict,
        _MetricsDict,
    )
    from great_expectations.validator.validator import (
        Validator,
    )


class MetricRetriever(abc.ABC):
    """A MetricRetriever is responsible for retrieving metrics for a batch of data. It is an ABC that contains base logic and
    methods share by both the ColumnDescriptiveMetricsMetricReceiver and MetricListMetricRetriver.
    """  # noqa: E501 # FIXME CoP

    def __init__(self, context: AbstractDataContext):
        self._context = context
        self._validator: Validator | None = None

    def get_validator(self, batch_request: BatchRequest) -> Validator:
        if self._validator is None:
            self._validator = self._context.get_validator(batch_request=batch_request)

        if isinstance(self._validator.active_batch, Batch):
            if self._validator.active_batch.data_asset.name != batch_request.data_asset_name:
                self._validator = self._context.get_validator(batch_request=batch_request)

        return self._validator

    @abc.abstractmethod
    def get_metrics(
        self,
        batch_request: BatchRequest,
        metric_list: Optional[List[MetricTypes]] = None,
    ) -> Sequence[Metric]:
        raise NotImplementedError

    def _generate_metric_id(self) -> uuid.UUID:
        return uuid.uuid4()

    def _get_metric_from_computed_metrics(
        self,
        metric_name: str | MetricTypes,
        computed_metrics: _MetricsDict,
        aborted_metrics: _AbortedMetricsInfoDict,
        metric_lookup_key: MetricConfigurationID | None = None,
    ) -> tuple[Any, MetricException | None]:
        # look up is done by string
        # TODO: update to be MetricTypes once MetricListMetricRetriever implementation is complete.
        if isinstance(metric_name, MetricTypes):
            metric_name = metric_name.value

        if metric_lookup_key is None:
            metric_lookup_key = MetricConfigurationID(
                metric_name,
                (),
                (),
            )
        value = None
        metric_exception = None
        if metric_lookup_key in computed_metrics:
            value = computed_metrics[metric_lookup_key]
        elif metric_lookup_key in aborted_metrics:
            exception = aborted_metrics[metric_lookup_key]
            exception_info = exception["exception_info"]
            exception_type = (
                "Unknown"  # Note: we currently only capture the message and traceback, not the type
            )
            if isinstance(exception_info, ExceptionInfo):
                exception_message = exception_info.exception_message
                metric_exception = MetricException(type=exception_type, message=exception_message)
        else:
            metric_exception = MetricException(
                type="Not found",
                message="Metric was not successfully computed but exception was not found.",
            )

        return value, metric_exception

    def _generate_table_metric_configurations(
        self, table_metric_names: list[str | MetricTypes]
    ) -> list[MetricConfiguration]:
        table_metric_configs = [
            MetricConfiguration(
                metric_name=metric_name, metric_domain_kwargs={}, metric_value_kwargs={}
            )
            for metric_name in table_metric_names
        ]
        return table_metric_configs

    def _compute_metrics(
        self, batch_request: BatchRequest, metric_configs: list[MetricConfiguration]
    ) -> tuple[str, _MetricsDict, _AbortedMetricsInfoDict]:
        validator = self.get_validator(batch_request=batch_request)
        # The runtime configuration catch_exceptions is explicitly set to True to catch exceptions
        # that are thrown when computing metrics. This is so we can capture the error for later
        # surfacing, and not have the entire metric run fail so that other metrics will still be
        # computed.
        (
            computed_metrics,
            aborted_metrics,
        ) = validator.compute_metrics(
            metric_configurations=metric_configs,
            runtime_configuration={"catch_exceptions": True},
        )
        assert isinstance(validator.active_batch, Batch), (
            f"validator.active_batch is type {type(validator.active_batch).__name__} "
            f"instead of type {Batch.__name__}"
        )
        batch_id = validator.active_batch.id
        return batch_id, computed_metrics, aborted_metrics

    def _get_columns_to_exclude(self, table_column_types: Metric) -> List[str]:
        """
        Excludes columns that are unsupported or missing type metadata
        """
        columns_to_skip: List[str] = []
        UNSUPPORTED_COLUMN_TYPES = ["TIME"]
        for column_type in table_column_types.value:
            if not column_type.get("type"):
                columns_to_skip.append(column_type["name"])
            if column_type.get("type") in UNSUPPORTED_COLUMN_TYPES:
                columns_to_skip.append(column_type["name"])
        return columns_to_skip

    def _get_numeric_column_names(
        self,
        batch_request: BatchRequest,
        exclude_column_names: List[str],
    ) -> list[str]:
        """Get the names of all numeric columns in the batch."""
        return self._get_column_names_for_semantic_types(
            batch_request=batch_request,
            include_semantic_types=[SemanticDomainTypes.NUMERIC],
            exclude_column_names=exclude_column_names,
        )

    def _get_timestamp_column_names(
        self,
        batch_request: BatchRequest,
        exclude_column_names: List[str],
    ) -> list[str]:
        """Get the names of all timestamp columns in the batch."""
        return self._get_column_names_for_semantic_types(
            batch_request=batch_request,
            include_semantic_types=[SemanticDomainTypes.DATETIME],
            exclude_column_names=exclude_column_names,
        )

    def _get_column_names_for_semantic_types(
        self,
        batch_request: BatchRequest,
        include_semantic_types: List[SemanticDomainTypes],
        exclude_column_names: List[str],
    ) -> list[str]:
        """Get the names of all columns matching semantic types in the batch."""
        validator = self.get_validator(batch_request=batch_request)
        column_filter = ColumnFilter(
            include_semantic_types=include_semantic_types,
            exclude_column_names=exclude_column_names,
        )
        assert isinstance(validator.active_batch, Batch), (
            f"validator.active_batch is type {type(validator.active_batch).__name__} "
            f"instead of type {Batch.__name__}"
        )
        column_names = column_filter.get_filtered_column_names(
            validator=validator,
        )
        return column_names

    def _get_table_metrics(
        self,
        batch_request: BatchRequest,
        metric_name: MetricTypes | str,
        metric_type: type[Metric],
    ) -> Metric:
        metric_configs = self._generate_table_metric_configurations([metric_name])
        batch_id, computed_metrics, aborted_metrics = self._compute_metrics(
            batch_request, metric_configs
        )
        value, exception = self._get_metric_from_computed_metrics(
            metric_name=metric_name,
            computed_metrics=computed_metrics,
            aborted_metrics=aborted_metrics,
        )
        return metric_type(
            batch_id=batch_id, metric_name=metric_name, value=value, exception=exception
        )

    def _get_column_metrics(
        self,
        batch_request: BatchRequest,
        column_list: List[str],
        column_metric_names: List[MetricTypes | str],
        column_metric_type: type[ColumnMetric[Any]],
    ) -> Sequence[Metric]:
        column_metric_configs = self._generate_column_metric_configurations(
            column_list, column_metric_names
        )
        batch_id, computed_metrics, aborted_metrics = self._compute_metrics(
            batch_request, column_metric_configs
        )

        # Convert computed_metrics
        ColumnMetric.update_forward_refs()
        metrics: list[Metric] = []

        for metric_name in column_metric_names:
            for column in column_list:
                metric_lookup_key = MetricConfigurationID(metric_name, f"column={column}", ())
                value, exception = self._get_metric_from_computed_metrics(
                    metric_name=metric_name,
                    metric_lookup_key=metric_lookup_key,
                    computed_metrics=computed_metrics,
                    aborted_metrics=aborted_metrics,
                )
                metrics.append(
                    column_metric_type(
                        batch_id=batch_id,
                        metric_name=metric_name,
                        column=column,
                        value=value,
                        exception=exception,
                    )
                )
        return metrics

    def _generate_column_metric_configurations(
        self, column_list: list[str], column_metric_names: list[str | MetricTypes]
    ) -> list[MetricConfiguration]:
        column_metric_configs: List[MetricConfiguration] = list()
        for metric_name in column_metric_names:
            for column in column_list:
                column_metric_configs.append(
                    MetricConfiguration(
                        metric_name=metric_name,
                        metric_domain_kwargs={"column": column},
                        metric_value_kwargs={},
                    )
                )
        return column_metric_configs

    def _get_all_column_names(self, metrics: Sequence[Metric]) -> List[str]:
        column_list: List[str] = []
        for metric in metrics:
            if metric.metric_name == MetricTypes.TABLE_COLUMNS:
                column_list = metric.value
        return column_list

    def _get_table_row_count(self, batch_request: BatchRequest) -> Metric:
        return self._get_table_metrics(
            batch_request=batch_request,
            metric_name=MetricTypes.TABLE_ROW_COUNT,
            metric_type=TableMetric[int],
        )

    def _get_table_columns(self, batch_request: BatchRequest) -> Metric:
        return self._get_table_metrics(
            batch_request=batch_request,
            metric_name=MetricTypes.TABLE_COLUMNS,
            metric_type=TableMetric[List[str]],
        )

    def _get_table_column_types(self, batch_request: BatchRequest) -> Metric:
        metric_name = MetricTypes.TABLE_COLUMN_TYPES

        metric_lookup_key = MetricConfigurationID(metric_name, (), ())
        table_metric_configs = self._generate_table_metric_configurations(
            table_metric_names=[metric_name]
        )
        batch_id, computed_metrics, aborted_metrics = self._compute_metrics(
            batch_request, table_metric_configs
        )
        value, exception = self._get_metric_from_computed_metrics(
            metric_name=metric_name,
            metric_lookup_key=metric_lookup_key,
            computed_metrics=computed_metrics,
            aborted_metrics=aborted_metrics,
        )
        raw_column_types: list[dict[str, Any]] = value
        # If type is not found, don't add empty type field. This can happen if our db introspection fails.  # noqa: E501 # FIXME CoP
        column_types_converted_to_str: list[dict[str, str]] = []
        for raw_column_type in raw_column_types:
            if raw_column_type.get("type"):
                column_types_converted_to_str.append(
                    {
                        "name": raw_column_type["name"],
                        "type": str(raw_column_type["type"]),
                    }
                )
            else:
                column_types_converted_to_str.append({"name": raw_column_type["name"]})

        return TableMetric[List[str]](
            batch_id=batch_id,
            metric_name=metric_name,
            value=column_types_converted_to_str,
            exception=exception,
        )



import pandas as pd

import great_expectations.expectations as gxe
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

COL_A = "col_a"
COL_B = "col_b"


DATA = pd.DataFrame({COL_A: [1, 2, None], COL_B: ["a", "b", None]})


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_success(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableRowCountToEqual(value=3)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableRowCountToEqual(value=2)
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING

import great_expectations as gx
from great_expectations.data_context.data_context.file_data_context import (
    FileDataContext,
)
from great_expectations.data_context.data_context.serializable_data_context import (
    SerializableDataContext,
)
from great_expectations.data_context.types.base import DataContextConfigDefaults

if TYPE_CHECKING:
    from great_expectations.data_context.data_context_variables import (
        DataContextVariables,
    )
    from great_expectations.data_context.store.datasource_store import DatasourceStore
    from great_expectations.data_context.store.store import Store
    from great_expectations.datasource.fluent.config import GxConfig

logger = logging.getLogger(__name__)


class FileMigrator:
    """Encapsulates any logic necessary to convert an existing context to a FileDataContext

    Only takes in the necessary dependencies for conversion:
        - context.stores
        - context._datasource_store
        - context.variables
        - context.fluent_config
    """

    def __init__(
        self,
        primary_stores: dict[str, Store],
        datasource_store: DatasourceStore,
        variables: DataContextVariables,
        fluent_config: GxConfig,
    ) -> None:
        self._primary_stores = primary_stores
        self._datasource_store = datasource_store
        self._variables = variables
        self._fluent_config = fluent_config

    def migrate(self) -> FileDataContext:
        """Migrate your in-memory Data Context to a file-backed one.

        Takes the following steps:
            1. Scaffolds filesystem
            2. Migrates primary stores (only creates default named stores)
            3. Migrates datasource store
            4. Migrates data docs sites (both physical files and config)
            5. Migrates fluent datasources

        Returns:
            A FileDataContext with an updated config to reflect the state of the current context.
        """
        target_context = self._scaffold_filesystem()
        self._migrate_primary_stores(
            target_stores=target_context.stores,
        )
        self._migrate_datasource_store(target_store=target_context._datasource_store)
        self._migrate_data_docs_sites(
            target_context=target_context,
        )
        self._migrate_fluent_datasources(target_context=target_context)

        # Re-init context to parse filesystem changes into config
        target_context = FileDataContext()
        print(f"Successfully migrated to {target_context.__class__.__name__}!")
        return target_context

    def _scaffold_filesystem(self) -> FileDataContext:
        path = pathlib.Path.cwd().absolute()
        target_context = gx.get_context(mode="file", project_root_dir=str(path))
        logger.info("Scaffolded necessary directories for a file-backed context")

        return target_context

    def _migrate_primary_stores(self, target_stores: dict[str, Store]) -> None:
        source_stores = self._primary_stores
        for name, source_store in source_stores.items():
            target_store = target_stores.get(name)
            if target_store:
                self._migrate_store(
                    store_name=name,
                    source_store=source_store,
                    target_store=target_store,
                )
            else:
                logger.warning(
                    f"Could not migrate the contents of store {name}; only default named stores are migrated"  # noqa: E501 # FIXME CoP
                )

    def _migrate_datasource_store(self, target_store: DatasourceStore) -> None:
        source_store = self._datasource_store
        self._migrate_store(
            store_name=DataContextConfigDefaults.DEFAULT_DATASOURCE_STORE_NAME.value,
            source_store=source_store,
            target_store=target_store,
        )

    def _migrate_store(self, store_name: str, source_store: Store, target_store: Store) -> None:
        logger.info(f"Migrating key-value pairs from {store_name} ({source_store.__class__}).")
        for key in source_store.list_keys():
            source_obj = source_store.get(key)
            target_store.add(key=key, value=source_obj)
            logger.info(f"Successfully migrated stored object saved with key {key}.")

    def _migrate_data_docs_sites(self, target_context: FileDataContext) -> None:
        target_root = pathlib.Path(target_context.root_directory)
        target_variables = target_context.variables
        source_configs = self._variables.data_docs_sites or {}

        self._migrate_data_docs_site_configs(
            target_root=target_root,
            source_configs=source_configs,
            target_variables=target_variables,
        )
        target_context.build_data_docs()

    def _migrate_fluent_datasources(self, target_context: FileDataContext) -> None:
        target_context.fluent_config = self._fluent_config
        target_context._save_project_config()

    def _migrate_data_docs_site_configs(
        self,
        source_configs: dict,
        target_root: pathlib.Path,
        target_variables: DataContextVariables,
    ):
        target_base_directory = target_root.joinpath(
            DataContextConfigDefaults.DEFAULT_DATA_DOCS_BASE_DIRECTORY_RELATIVE_NAME.value
        )

        updated_data_docs_config = {}
        for site_name, site_config in source_configs.items():
            updated_site_config = self._migrate_data_docs_site_config(
                site_name=site_name,
                site_config=site_config,
                target_base_directory=target_base_directory,
            )
            updated_data_docs_config[site_name] = updated_site_config

        # If no sites to migrate, don't touch config defaults
        if updated_data_docs_config:
            target_variables.data_docs_sites = updated_data_docs_config
            target_variables.save()

    def _migrate_data_docs_site_config(
        self, site_name: str, site_config: dict, target_base_directory: pathlib.Path
    ) -> dict:
        absolute_site_path = target_base_directory.joinpath(site_name)
        project_root = pathlib.Path.cwd().joinpath(SerializableDataContext.GX_DIR)
        relative_site_path = absolute_site_path.relative_to(project_root)

        updated_config = site_config
        updated_config["store_backend"]["base_directory"] = str(relative_site_path)

        return updated_config



from __future__ import annotations

import pathlib
from abc import ABC
from functools import singledispatchmethod
from typing import TYPE_CHECKING, Generic, Optional

from great_expectations import exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.partitioners import (
    ColumnPartitioner,
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
)
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.batch_identifier_util import make_batch_identifier
from great_expectations.datasource.fluent.constants import _DATA_CONNECTOR_NAME
from great_expectations.datasource.fluent.data_asset.path.dataframe_partitioners import (
    DataframePartitioner,
    DataframePartitionerDaily,
    DataframePartitionerMonthly,
    DataframePartitionerYearly,
)
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)
from great_expectations.datasource.fluent.data_connector import FILE_PATH_BATCH_SPEC_KEY
from great_expectations.datasource.fluent.interfaces import DatasourceT, PartitionerSortingProtocol

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr
    from great_expectations.core.batch_definition import BatchDefinition
    from great_expectations.datasource.fluent import BatchParameters
    from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice


@public_api
class DirectoryDataAsset(PathDataAsset[DatasourceT, ColumnPartitioner], Generic[DatasourceT], ABC):
    """Base class for PathDataAssets which batch by combining the contents of a directory."""

    data_directory: pathlib.Path

    @public_api
    def add_batch_definition_daily(self, name: str, column: str) -> BatchDefinition:
        """
        Add a BatchDefinition, which creates a single Batch for each day in the directory.

        Args:
            name: Name of the Batch Definition.
            column: Column to partition on.

        Returns:
            A BatchDefinition that is partitioned daily.
        """
        # todo: test column
        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerDaily(
                method_name="partition_on_year_and_month_and_day", column_name=column
            ),
        )

    @public_api
    def add_batch_definition_monthly(self, name: str, column: str) -> BatchDefinition:
        """
        Add a BatchDefinition which creates a single batch for each month in the directory.

        Args:
            name: Name of the Batch Definition.
            column: Column to partition on.

        Returns:
            A BatchDefinition that is partitioned monthly.
        """
        # todo: test column
        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerMonthly(
                method_name="partition_on_year_and_month", column_name=column
            ),
        )

    @public_api
    def add_batch_definition_yearly(self, name: str, column: str) -> BatchDefinition:
        """
        Add a BatchDefinition which creates a single batch for each year in the directory.

        Args:
            name: Name of the Batch Definition.
            column: Column to partition on.

        Returns:
            A BatchDefinition that is partitioned yearly.
        """
        # todo: test column
        return self.add_batch_definition(
            name=name,
            partitioner=ColumnPartitionerYearly(
                method_name="partition_on_year", column_name=column
            ),
        )

    @public_api
    def add_batch_definition_whole_directory(self, name: str) -> BatchDefinition:
        """Add a BatchDefinition which creates a single batch for the entire directory."""
        return self.add_batch_definition(name=name, partitioner=None)

    @override
    def _get_batch_definition_list(
        self, batch_request: BatchRequest
    ) -> list[LegacyBatchDefinition]:
        """Generate a batch definition list from a given batch request.

        Args:
            batch_request: Batch request used to generate batch definitions.

        Returns:
            List of a single batch definition.
        """
        if batch_request.partitioner:
            # Currently non-sql asset partitioners do not introspect the datasource for available
            # batches and only return a single batch based on specified batch_identifiers.
            batch_identifiers = batch_request.options
            if not batch_identifiers.get("path"):
                batch_identifiers["path"] = self.data_directory

            batch_definition = LegacyBatchDefinition(
                datasource_name=self._data_connector.datasource_name,
                data_connector_name=_DATA_CONNECTOR_NAME,
                data_asset_name=self._data_connector.data_asset_name,
                batch_identifiers=make_batch_identifier(batch_identifiers),
            )
            batch_definition_list = [batch_definition]
        else:
            batch_definition_list = self._data_connector.get_batch_definition_list(
                batch_request=batch_request
            )
        return batch_definition_list

    @singledispatchmethod
    def _get_dataframe_partitioner(self, partitioner) -> Optional[DataframePartitioner]: ...

    @_get_dataframe_partitioner.register
    def _(self, partitioner: ColumnPartitionerYearly) -> DataframePartitionerYearly:
        return DataframePartitionerYearly(**partitioner.dict(exclude={"param_names"}))

    @_get_dataframe_partitioner.register
    def _(self, partitioner: ColumnPartitionerMonthly) -> DataframePartitionerMonthly:
        return DataframePartitionerMonthly(**partitioner.dict(exclude={"param_names"}))

    @_get_dataframe_partitioner.register
    def _(self, partitioner: ColumnPartitionerDaily) -> DataframePartitionerDaily:
        return DataframePartitionerDaily(**partitioner.dict(exclude={"param_names"}))

    @_get_dataframe_partitioner.register
    def _(self, partitioner: None) -> None:
        return None

    @override
    def _get_reader_options_include(self) -> set[str]:
        return {
            "data_directory",
        }

    @override
    def get_batch_parameters_keys(
        self,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> tuple[str, ...]:
        option_keys: tuple[str, ...] = (FILE_PATH_BATCH_SPEC_KEY,)
        dataframe_partitioner = self._get_dataframe_partitioner(partitioner)
        if dataframe_partitioner:
            option_keys += tuple(dataframe_partitioner.param_names)
        return option_keys

    @override
    def get_whole_directory_path_override(
        self,
    ) -> PathStr:
        return self.data_directory

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> BatchRequest:
        if options is not None and not self._batch_parameters_are_valid(
            options=options,
            partitioner=partitioner,
        ):
            allowed_keys = set(self.get_batch_parameters_keys(partitioner=partitioner))
            actual_keys = set(options.keys())
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "Batch parameters should only contain keys from the following set:\n"
                f"{allowed_keys}\nbut your specified keys contain\n"
                f"{actual_keys.difference(allowed_keys)}\nwhich is not valid.\n"
            )

        return BatchRequest(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            options=options or {},
            batch_slice=batch_slice,
            partitioner=partitioner,
        )

    @override
    def _batch_spec_options_from_batch_request(self, batch_request: BatchRequest) -> dict:
        """Build a set of options for use in a batch spec from a batch request.

        Args:
            batch_request: Batch request to use to generate options.

        Returns:
            Dictionary containing batch spec options.
        """
        get_reader_options_include: set[str] | None = self._get_reader_options_include()
        if not get_reader_options_include:
            # Set to None if empty set to include any additional `extra_kwargs` passed to `add_*_asset`  # noqa: E501 # FIXME CoP
            get_reader_options_include = None
        batch_spec_options = {
            "reader_method": self._get_reader_method(),
            "reader_options": self.dict(
                include=get_reader_options_include,
                exclude=self._EXCLUDE_FROM_READER_OPTIONS,
                exclude_unset=True,
                by_alias=True,
                config_provider=self._datasource._config_provider,
            ),
        }

        partitioner_parameters = self._get_partitioner_parameters(batch_request=batch_request)
        if partitioner_parameters:
            batch_spec_options.update(partitioner_parameters)

        return batch_spec_options

    def _get_partitioner_parameters(self, batch_request: BatchRequest) -> Optional[dict]:
        """If a partitioner is present, add its configuration to batch parameters."""
        partitioner: Optional[DataframePartitioner] = self._get_dataframe_partitioner(
            batch_request.partitioner
        )
        if not partitioner:
            return None
        batch_identifiers = partitioner.batch_parameters_to_batch_spec_kwarg_identifiers(
            parameters=batch_request.options
        )
        return {
            "partitioner_method": partitioner.method_name,
            "partitioner_kwargs": {
                **partitioner.partitioner_method_kwargs(),
                "batch_identifiers": batch_identifiers,
            },
        }

    @override
    def _get_sortable_partitioner(
        self, partitioner: Optional[ColumnPartitioner]
    ) -> Optional[PartitionerSortingProtocol]:
        # DirectoryAssets can only ever return a single batch, so they do not require sorting.
        return None



from __future__ import annotations

import copy
import inspect
import logging
import pathlib
import re
from dataclasses import dataclass
from pprint import pformat as pf
from typing import TYPE_CHECKING, Any, Optional, Type

import pytest
from pytest import MonkeyPatch, param

import great_expectations.exceptions as ge_exceptions
import great_expectations.execution_engine.pandas_execution_engine
from great_expectations.compatibility import pydantic
from great_expectations.core.partitioners import FileNamePartitionerMonthly
from great_expectations.datasource.fluent import PandasFilesystemDatasource
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    CSVAsset,
    JSONAsset,
)
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)
from great_expectations.datasource.fluent.data_connector import (
    FilesystemDataConnector,
)
from great_expectations.datasource.fluent.dynamic_pandas import PANDAS_VERSION
from great_expectations.datasource.fluent.interfaces import TestConnectionError
from great_expectations.datasource.fluent.sources import _get_field_details
from great_expectations.exceptions.exceptions import NoAvailableBatchesError

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.datasource.fluent.interfaces import (
        BatchMetadata,
        BatchSlice,
    )

logger = logging.getLogger(__file__)

# apply markers to entire test module
pytestmark = [
    pytest.mark.skipif(
        PANDAS_VERSION < 1.2, reason=f"Fluent pandas not supported on {PANDAS_VERSION}"
    )
]


class SpyInterrupt(RuntimeError):
    """
    Exception that may be raised to interrupt the control flow of the program
    when a spy has already captured everything needed.
    """


@pytest.fixture
def capture_reader_fn_params(monkeypatch: MonkeyPatch):
    """
    Capture the `reader_options` arguments being passed to the `PandasExecutionEngine`.

    Note this fixture is heavily reliant on the implementation details of `PandasExecutionEngine`,
    should this change this fixture will need to change.
    """
    captured_args: list[list] = []
    captured_kwargs: list[dict[str, Any]] = []

    def reader_fn_spy(*args, **kwargs):
        logging.info(f"reader_fn_spy() called with...\n{args}\n{kwargs}")
        captured_args.append(args)
        captured_kwargs.append(kwargs)
        raise SpyInterrupt("Reader options have been captured")

    monkeypatch.setattr(
        great_expectations.execution_engine.pandas_execution_engine.PandasExecutionEngine,
        "_get_reader_fn",
        lambda *_: reader_fn_spy,
        raising=True,
    )

    yield captured_args, captured_kwargs


@pytest.mark.unit
class TestDynamicPandasAssets:
    @pytest.mark.parametrize(
        "method_name",
        [
            param("read_clipboard", marks=pytest.mark.xfail(reason="not path based")),
            param("read_csv"),
            param("read_excel"),
            param("read_feather"),
            param("read_fwf"),
            param("read_gbq", marks=pytest.mark.xfail(reason="not path based")),
            param("read_hdf"),
            param("read_html"),
            param("read_json"),
            param("read_orc"),
            param("read_parquet"),
            param("read_pickle"),
            param("read_sas"),
            param("read_spss"),
            param(
                "read_sql",
                marks=pytest.mark.xfail(reason="name conflict & not path based"),
            ),
            param(
                "read_sql_query",
                marks=pytest.mark.xfail(
                    reason="type name logic expects 'sqltable' & not path based"
                ),
            ),
            param(
                "read_sql_table",
                marks=pytest.mark.xfail(
                    reason="type name logic expects 'sqltable' & not path based"
                ),
            ),
            param("read_stata"),
            param(
                "read_table",
                marks=pytest.mark.xfail(reason="name conflict & not path based"),
            ),
            param(
                "read_xml",
                marks=pytest.mark.skipif(
                    PANDAS_VERSION < 1.3,
                    reason=f"read_xml does not exist on {PANDAS_VERSION} ",
                ),
            ),
        ],
    )
    def test_data_asset_defined_for_io_read_method(self, method_name: str):
        _, type_name = method_name.split("read_")
        assert type_name

        asset_class_names: set[str] = {
            t.__name__.lower().split("asset")[0] for t in PandasFilesystemDatasource.asset_types
        }
        print(asset_class_names)

        assert type_name in asset_class_names

    @pytest.mark.parametrize("asset_class", PandasFilesystemDatasource.asset_types)
    def test_add_asset_method_exists_and_is_functional(self, asset_class: Type[PathDataAsset]):
        type_name: str = _get_field_details(asset_class, "type").default_value
        method_name: str = f"add_{type_name}_asset"

        print(f"{method_name}() -> {asset_class.__name__}")

        assert method_name in PandasFilesystemDatasource.__dict__

        ds = PandasFilesystemDatasource(
            name="ds_for_testing_add_asset_methods",
            base_directory=pathlib.Path.cwd(),
        )
        method = getattr(ds, method_name)

        with pytest.raises(pydantic.ValidationError) as exc_info:
            method(
                f"{asset_class.__name__}_add_asset_test",
                batching_regex="great_expectations",
                _invalid_key="foobar",
            )
        # importantly check that the method creates (or attempts to create) the intended asset
        assert exc_info.value.model == asset_class

    @pytest.mark.parametrize("asset_class", PandasFilesystemDatasource.asset_types)
    def test_add_asset_method_signature(self, asset_class: Type[PathDataAsset]):
        type_name: str = _get_field_details(asset_class, "type").default_value
        method_name: str = f"add_{type_name}_asset"

        ds = PandasFilesystemDatasource(
            name="ds_for_testing_add_asset_methods",
            base_directory=pathlib.Path.cwd(),
        )
        method = getattr(ds, method_name)

        add_asset_method_sig: inspect.Signature = inspect.signature(method)
        print(f"\t{method_name}()\n{add_asset_method_sig}\n")

        asset_class_init_sig: inspect.Signature = inspect.signature(asset_class)
        print(f"\t{asset_class.__name__}\n{asset_class_init_sig}\n")

        for i, param_name in enumerate(asset_class_init_sig.parameters):
            print(f"{i} {param_name} ", end="")

            if param_name == "type":
                assert param_name not in add_asset_method_sig.parameters, (
                    "type should not be part of the `add_<TYPE>_asset` method"
                )
                print("⏩")
                continue

            assert param_name in add_asset_method_sig.parameters
            print("✅")

    @pytest.mark.parametrize(
        ["asset_model", "extra_kwargs"],
        [
            (CSVAsset, {"sep": "|", "names": ["col1", "col2", "col3"]}),
            (JSONAsset, {"orient": "records", "convert_dates": True}),
        ],
    )
    def test_data_asset_defaults(
        self,
        asset_model: Type[PathDataAsset],
        extra_kwargs: dict,
    ):
        """
        Test that an asset dictionary can be dumped with only the original passed keys
        present.
        """
        kwargs: dict[str, Any] = {
            "name": "test",
        }
        kwargs.update(extra_kwargs)
        print(f"extra_kwargs\n{pf(extra_kwargs)}")
        asset_instance = asset_model(**kwargs)
        assert asset_instance.dict(exclude={"type"}) == kwargs

    @pytest.mark.parametrize(
        "extra_kwargs",
        [
            {"sep": "|", "decimal": ","},
            {"usecols": [0, 1, 2], "names": ["foo", "bar"]},
            {"dtype": {"col_1": "Int64"}},
        ],
    )
    def test_data_asset_reader_options_passthrough(
        self,
        empty_data_context: AbstractDataContext,
        csv_path: pathlib.Path,
        capture_reader_fn_params: tuple[list[list], list[dict]],
        extra_kwargs: dict,
    ):
        batch_request = (
            empty_data_context.data_sources.add_pandas_filesystem(  # .build_batch_request
                "my_pandas",
                base_directory=csv_path,
            )
            .add_csv_asset(
                "my_csv",
                **extra_kwargs,
            )
            .build_batch_request(
                {"year": "2018"},
                partitioner=FileNamePartitionerMonthly(
                    regex=re.compile(
                        r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
                    )
                ),
            )
        )
        with pytest.raises(SpyInterrupt):
            empty_data_context.get_validator(batch_request=batch_request)

        captured_args, captured_kwargs = capture_reader_fn_params
        print(f"positional args:\n{pf(captured_args[-1])}\n")
        print(f"keyword args:\n{pf(captured_kwargs[-1])}")

        assert captured_kwargs[-1] == extra_kwargs


@pytest.mark.unit
def test_construct_pandas_filesystem_datasource(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    assert pandas_filesystem_datasource.name == "pandas_filesystem_datasource"


@pytest.mark.unit
def test_add_csv_asset_to_datasource(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
def test_add_csv_asset_with_batching_regex_to_datasource(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    assert asset.name == "csv_asset"


@pytest.mark.unit
def test_invalid_connect_options(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    with pytest.raises(pydantic.ValidationError) as exc_info:
        pandas_filesystem_datasource.add_csv_asset(  # type: ignore[call-arg] # FIXME CoP
            name="csv_asset",
            glob_foobar="invalid",
        )

    error_dicts = exc_info.value.errors()
    print(pf(error_dicts))
    assert error_dicts == [
        {
            "loc": ("glob_foobar",),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        }
    ]


@pytest.mark.unit
@pytest.mark.parametrize(
    ["glob_directive", "expected_error"],
    [
        ({"invalid", "type"}, pydantic.ValidationError),
        ("not_a_dir/*.csv", TestConnectionError),
    ],
)
def test_invalid_connect_options_value(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
    glob_directive,
    expected_error: Type[Exception],
):
    with pytest.raises(expected_error) as exc_info:
        pandas_filesystem_datasource.add_csv_asset(
            name="csv_asset",
            glob_directive=glob_directive,
        )

    print(f"Exception raised:\n\t{exc_info.value!r}")

    if isinstance(exc_info.value, pydantic.ValidationError):
        error_dicts = exc_info.value.errors()
        print(pf(error_dicts))
        assert error_dicts == [
            {
                "loc": ("glob_directive",),
                "msg": "str type expected",
                "type": "type_error.str",
            }
        ]


@pytest.mark.unit
@pytest.mark.parametrize(
    "connect_options",
    [
        param({"glob_directive": "**/*"}, id="glob **/*"),
        param({"glob_directive": "**/*.csv"}, id="glob **/*.csv"),
        param({}, id="default connect options"),
    ],
)
def test_asset_connect_options_in_repr(
    pandas_filesystem_datasource: PandasFilesystemDatasource, connect_options: dict
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        **connect_options,
    )
    asset_repr = repr(asset)
    print(asset_repr)

    if connect_options:
        assert "glob_directive" in asset_repr
        assert connect_options["glob_directive"] in asset_repr
    else:
        # if no connect options are provided the defaults should be used and should not
        # be part of any serialization. repr == asset.yaml()
        assert "glob_directive" not in asset_repr


@pytest.mark.unit
def test_csv_asset_with_batching_regex_named_parameters(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    batching_regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=batching_regex)
    options = asset.get_batch_parameters_keys(partitioner=batch_def.partitioner)
    assert options == ("path", "year", "month")


@pytest.mark.unit
def test_csv_asset_with_non_string_batching_regex_named_parameters(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    with pytest.raises(ge_exceptions.InvalidBatchRequestError):
        # year is an int which will raise an error
        asset.build_batch_request({"year": 2018, "month": "04"})


@pytest.mark.unit
def test_get_batch_list_from_fully_specified_batch_request(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=regex)
    batch_parameters = {"year": "2018", "month": "04"}
    batch = batch_def.get_batch(batch_parameters=batch_parameters)
    assert batch.batch_request.datasource_name == pandas_filesystem_datasource.name
    assert batch.batch_request.data_asset_name == asset.name

    path = "yellow_tripdata_sample_2018-04.csv"
    assert batch.batch_request.options == {"path": path, "year": "2018", "month": "04"}
    assert batch.metadata == {"path": path, "year": "2018", "month": "04"}

    assert batch.id == "pandas_filesystem_datasource-csv_asset-year_2018-month_04"


@pytest.mark.unit
@pytest.mark.parametrize(
    "year,month,path,batch_count",
    [
        ("2018", "04", "yellow_tripdata_sample_2018-04.csv", 1),
        ("2018", None, None, 12),
        (None, "04", None, 3),
        (None, "03", "yellow_tripdata_sample_2018-04.csv", 0),
    ],
)
def test_get_batch_identifiers_list_count(
    year: Optional[str],
    month: Optional[str],
    path: Optional[str],
    batch_count: int,
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    request = asset.build_batch_request(
        {"year": year, "month": month, "path": path},
        partitioner=FileNamePartitionerMonthly(
            regex=re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
        ),
    )
    batch_identifier_list = asset.get_batch_identifiers_list(request)
    assert len(batch_identifier_list) == batch_count


@pytest.mark.unit
def test_get_batch_identifiers_list_from_partially_specified_batch_request(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    # Verify test directory has files that don't match what we will query for
    file_name: PathStr
    all_files: list[str] = [
        file_name.stem
        for file_name in list(pathlib.Path(pandas_filesystem_datasource.base_directory).iterdir())
    ]
    # assert there are files that are not csv files
    assert any(not file_name.endswith("csv") for file_name in all_files)
    # assert there are 12 files from 2018
    files_for_2018 = [file_name for file_name in all_files if file_name.find("2018") >= 0]
    assert len(files_for_2018) == 12

    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    request = asset.build_batch_request(
        {"year": "2018"},
        partitioner=FileNamePartitionerMonthly(
            regex=re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
        ),
    )
    batches = asset.get_batch_identifiers_list(request)
    assert (len(batches)) == 12
    batch_filenames = [pathlib.Path(batch["path"]).stem for batch in batches]
    assert set(files_for_2018) == set(batch_filenames)

    @dataclass(frozen=True)
    class YearMonth:
        year: str
        month: str

    expected_year_month = {YearMonth(year="2018", month=format(m, "02d")) for m in range(1, 13)}
    batch_year_month = {YearMonth(year=batch["year"], month=batch["month"]) for batch in batches}
    assert expected_year_month == batch_year_month


@pytest.mark.unit
@pytest.mark.parametrize(
    "batch_slice, expected_batch_count",
    [
        ("[-3:]", 3),
        ("[5:9]", 4),
        ("[:10:2]", 5),
        (slice(-3, None), 3),
        (slice(5, 9), 4),
        (slice(0, 10, 2), 5),
        ("-5", 1),
        ("-1", 1),
        (11, 1),
        (0, 1),
        ([3], 1),
        (None, 12),
        ("", 12),
    ],
)
def test_pandas_slice_batch_count(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
    batch_slice: BatchSlice,
    expected_batch_count: int,
) -> None:
    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
    )
    batch_request = asset.build_batch_request(
        options={"year": "2019"},
        batch_slice=batch_slice,
        partitioner=FileNamePartitionerMonthly(
            regex=re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
        ),
    )
    batch_identifiers_list = asset.get_batch_identifiers_list(batch_request=batch_request)
    assert len(batch_identifiers_list) == expected_batch_count


def bad_batching_regex_config(
    csv_path: pathlib.Path,
) -> tuple[re.Pattern, TestConnectionError]:
    batching_regex = re.compile(r"green_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
    test_connection_error = TestConnectionError(
        "No file at base_directory path "
        f'"{csv_path.resolve()}" matched regular expressions pattern '
        f'"{batching_regex.pattern}" and/or glob_directive "**/*" for '
        'DataAsset "csv_asset".'
    )
    return batching_regex, test_connection_error


@pytest.fixture(params=[bad_batching_regex_config])
def datasource_test_connection_error_messages(
    csv_path: pathlib.Path,
    pandas_filesystem_datasource: PandasFilesystemDatasource,
    request,
) -> tuple[PandasFilesystemDatasource, TestConnectionError]:
    _, test_connection_error = request.param(csv_path=csv_path)
    csv_asset = CSVAsset(  # type: ignore[call-arg] # FIXME CoP
        name="csv_asset",
    )
    csv_asset._datasource = pandas_filesystem_datasource
    pandas_filesystem_datasource.assets = [
        csv_asset,
    ]
    csv_asset._data_connector = FilesystemDataConnector(
        datasource_name=pandas_filesystem_datasource.name,
        data_asset_name=csv_asset.name,
        base_directory=pandas_filesystem_datasource.base_directory,
        data_context_root_directory=pandas_filesystem_datasource.data_context_root_directory,
    )
    csv_asset._test_connection_error_message = test_connection_error
    return pandas_filesystem_datasource, test_connection_error


@pytest.mark.timeout(
    5,  # deepcopy operation can be slow. Try to eliminate it in the future.
)
@pytest.mark.unit
def test_csv_asset_batch_metadata(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
):
    my_config_variables = {"pipeline_filename": __file__}
    pandas_filesystem_datasource._data_context.config_variables.update(  # type: ignore[union-attr] # `_data_context`
        my_config_variables
    )

    asset_specified_metadata = {
        "pipeline_name": "my_pipeline",
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
    }

    asset = pandas_filesystem_datasource.add_csv_asset(
        name="csv_asset",
        batch_metadata=asset_specified_metadata,
    )
    assert asset.batch_metadata == asset_specified_metadata

    batch_request = asset.build_batch_request(
        partitioner=FileNamePartitionerMonthly(
            regex=re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")
        )
    )

    batch = pandas_filesystem_datasource.get_batch(batch_request)

    substituted_batch_metadata: BatchMetadata = copy.deepcopy(asset_specified_metadata)
    substituted_batch_metadata.update(
        {
            "no_curly_pipeline_filename": __file__,
            "curly_pipeline_filename": __file__,
        }
    )

    actual_metadata = copy.deepcopy(batch.metadata)

    actual_metadata.pop("path")
    actual_metadata.pop("year")
    actual_metadata.pop("month")

    assert len(actual_metadata)
    assert actual_metadata == substituted_batch_metadata


@pytest.mark.parametrize(
    ("sort_ascending", "expected_metadata"),
    [
        (True, {"year": "2020", "month": "12", "path": "yellow_tripdata_sample_2020-12.csv"}),
        (False, {"year": "2018", "month": "01", "path": "yellow_tripdata_sample_2018-01.csv"}),
    ],
)
@pytest.mark.unit
def test_get_batch_respects_order_ascending(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
    sort_ascending: bool,
    expected_metadata: dict,
) -> None:
    asset = pandas_filesystem_datasource.add_csv_asset(name="csv_asset")
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(
        name="batch def", regex=regex, sort_ascending=sort_ascending
    )
    batch = batch_def.get_batch()
    assert batch.metadata == expected_metadata


@pytest.mark.unit
def test_raises_if_no_matching_batches(
    pandas_filesystem_datasource: PandasFilesystemDatasource,
) -> None:
    asset = pandas_filesystem_datasource.add_csv_asset(name="csv_asset")
    regex = r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv"
    batch_def = asset.add_batch_definition_monthly(name="batch def", regex=regex)
    with pytest.raises(NoAvailableBatchesError):
        batch_def.get_batch(batch_parameters={"year": "1995", "month": "01"})



from __future__ import annotations

from great_expectations.compatibility.typing_extensions import override

"""Serializer class interface definition.

Serializers determine how to write an object to disk, json, etc.
A serializer comprises the object destination and name e.g. YAMLReadyDictMyModelConfigSerializer.
A base implementation (DictConfigSerializer) is provided if no modification needs to be included for the specific object / destination pair.

Typical usage example:

config = ModelConfig(...)
serializer = DictConfigSerializer(schema=modelConfigSchema)
serialized_value = serializer.serialize(config)
"""  # noqa: E501 # FIXME CoP

import abc
from typing import TYPE_CHECKING

from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP

if TYPE_CHECKING:
    from marshmallow import Schema

    from great_expectations.core.configuration import AbstractConfig


class AbstractConfigSerializer(abc.ABC):
    """Serializer interface.

    Note: When mypy coverage is enhanced further, this Abstract class can be replaced with a Protocol.
    """  # noqa: E501 # FIXME CoP

    def __init__(self, schema: Schema) -> None:
        """
        Args:
            schema: Marshmallow schema defining raw serialized version of object.
        """
        self.schema = schema

    @abc.abstractmethod
    def serialize(self, obj: AbstractConfig) -> dict:
        """Serialize to serializer specific data type.

        Note, specific return type to be implemented in subclasses.

        Args:
            obj: Object to serialize.

        Returns:
            Representation of object in serializer specific data type.
        """
        raise NotImplementedError


class DictConfigSerializer(AbstractConfigSerializer):
    @override
    def serialize(self, obj: AbstractConfig) -> dict:
        """Serialize to Python dictionary.

        This is typically the default implementation used in can be overridden in subclasses.

        Args:
            obj: Object to serialize.

        Returns:
            Representation of object as a Python dictionary using the defined Marshmallow schema.
        """
        return self.schema.dump(obj)


class JsonConfigSerializer(AbstractConfigSerializer):
    @override
    def serialize(self, obj: AbstractConfig) -> dict:
        """Serialize config to json dict.

        Args:
            obj: AbstractConfig object to serialize.

        Returns:
            Representation of object as a dict suitable for serializing to json.
        """

        config: dict = self.schema.dump(obj)

        json_serializable_dict: dict = convert_to_json_serializable(data=config)

        return json_serializable_dict



import json
from pathlib import Path

import jsonschema
import pytest
from jsonschema import Draft7Validator

from great_expectations.expectations import core
from great_expectations.expectations.core import schemas
from great_expectations.expectations.expectation import MetaExpectation

expectation_dictionary = dict(core.__dict__)


@pytest.fixture
def safer_draft_7_validator() -> type[Draft7Validator]:
    validator = Draft7Validator
    validator.META_SCHEMA = {
        **Draft7Validator.META_SCHEMA,
        # this ensures that only specified properties are used (e.g. multipleOf, not multiple_of)
        # otherwise, the spec says unspecified properties should be ignored
        "additionalProperties": False,
    }
    return validator


@pytest.mark.unit
def test_all_core_model_schemas_are_serializable():
    all_models = [
        expectation
        for expectation in expectation_dictionary.values()
        if isinstance(expectation, MetaExpectation)
    ]
    # are they still there?
    assert len(all_models) > 50
    for model in all_models:
        model.schema_json()


@pytest.mark.filesystem  # ~4s
def test_schemas_updated():
    all_models = {
        cls_name: expectation
        for cls_name, expectation in expectation_dictionary.items()
        if isinstance(expectation, MetaExpectation)
    }
    schema_file_paths = Path(schemas.__file__).parent.glob("*.json")
    all_schemas = {file_path.stem: file_path.read_text() for file_path in schema_file_paths}
    for cls_name, schema in all_schemas.items():
        # converting to dicts for easier comparision on failure
        new_schema = json.loads(all_models[cls_name].schema_json())
        old_schema = json.loads(schema)
        assert new_schema == old_schema, "json schemas not updated, run `invoke schemas --sync`"


@pytest.mark.unit
def test_schemas_valid_spec(safer_draft_7_validator: type[Draft7Validator]):
    # https://json-schema.org/draft-07
    # https://jsonforms.io/api/core/interfaces/jsonschema7
    schema_file_paths = Path(schemas.__file__).parent.glob("*.json")
    for file_path in schema_file_paths:
        with open(file_path) as schema_file:
            try:
                safer_draft_7_validator.check_schema(json.load(schema_file))
            except jsonschema.exceptions.SchemaError as e:
                raise AssertionError(
                    f"Invalid json schema for `{file_path.name}`: {e.message}"
                ) from e



from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnValuesNotMatchRegexValuesResult(MetricResult[list[str]]): ...


class ColumnValuesNotMatchRegexValues(ColumnMetric[ColumnValuesNotMatchRegexValuesResult]):
    """Count of values in a column that do not match a regex"""

    name = "column_values.not_match_regex_values"
    regex: str
    limit: int = 20



from abc import ABC, abstractmethod
from typing import Generic, Iterable, TypeVar

T = TypeVar("T")


class Factory(ABC, Generic[T]):
    """
    Responsible for basic CRUD operations on collections of GX domain objects.
    """

    @abstractmethod
    def add(self, obj: T) -> T:
        pass

    @abstractmethod
    def delete(self, name: str) -> None:
        pass

    @abstractmethod
    def get(self, name: str) -> T:
        pass

    @abstractmethod
    def all(self) -> Iterable[T]:
        pass

    @abstractmethod
    def add_or_update(self, obj: T) -> T:
        pass



from great_expectations.compatibility.pydantic import StrictStr
from great_expectations.metrics.column import ColumnMetric
from great_expectations.metrics.metric_results import MetricResult


class ColumnValuesNotMatchRegexCountResult(MetricResult[int]): ...


class ColumnValuesNotMatchRegexCount(ColumnMetric[ColumnValuesNotMatchRegexCountResult]):
    """Count of values in a column that do not match a regex"""

    name = "column_values.not_match_regex.count"

    regex: StrictStr
